# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

from astrometry.util.util import *
from astrometry.util.starutil_numpy import *
from numpy import *
from pylab import *

def cosdeg(y):
    return cos(deg2rad(y))

def plot_healpix_on_radec(nside=2, cosdec=False, fn=None, textkws=None):

    def ycoord(y):
        if cosdec:
            return sin(deg2rad(y))
        else:
            return y

    clf()
    
    # RA,Dec grid lines.
    for x in range(-50,410,10):
        axvline(x=x, lw=0.5, ls='-', color='0.8')
    for y in range(-90,100,10):
        axhline(y=ycoord(y), lw=0.5, ls='-', color='0.8')

    if textkws is None:
        textkws = {}

    if not 'horizontalalignment' in textkws:
        textkws['horizontalalignment'] = 'center'
    if not 'verticalalignment' in textkws:
        textkws['verticalalignment'] = 'center'

    for hp in range(12 * nside**2):
        (racen,deccen) = healpix_to_radec(hp, nside, 0.5, 0.5)
        text(racen, ycoord(deccen), '%i' % hp, **textkws)

        nsteps = 20
        stepvals = [x/float(nsteps) for x in range(nsteps+1)]
        rstepvals = [1-x for x in stepvals]
        ones = [1.0]*(nsteps+1)
        zeros = [0.0]*(nsteps+1)

        x = stepvals + ones + rstepvals + zeros
        y = zeros + stepvals + ones + rstepvals

        edges = [healpix_to_radec(hp, nside, xx, yy) for (xx,yy) in zip(x,y)]
        ras = [ra for (ra,dec) in edges]
        decs = [dec for (ra,dec) in edges]

        if max(ras) > 270 and min(ras) < 90:
            # wrap-around!
            ras1 = []
            ras2 = []
            for ra in ras:
                if ra < 180:
                    ras1.append(ra)
                    ras2.append(ra+360)
                else:
                    ras1.append(ra-360)
                    ras2.append(ra)
            plot(ras1, ycoord(decs), 'k--')
            plot(ras2, ycoord(decs), 'k-')
            if racen > 180:
                text(racen-360, ycoord(deccen), '%i' % hp, **textkws)
            else:
                text(racen+360, ycoord(deccen), '%i' % hp, **textkws)

        else:
            plot(ras, ycoord(decs), 'k-')

    xlabel('RA (deg)')
    ylabel('Dec (deg)')
    ylim(ycoord(-90), ycoord(90))
    xlim(-40,400)
    yt = arange(-90, 100, 10)
    yticks(ycoord(yt), ['%g'%y for y in yt])
    title('Healpixes (nside=%i)' % nside)
    if fn is None:
        fn = 'healpix-%i.png' % nside
    savefig(fn)

def plot_radec_on_healpix(nside=1, bighp=1):
    clf()
    xlabel('Healpix x')
    ylabel('Healpix y')

    (ralo,rahi,declo,dechi) = healpix_radec_bounds(bighp, 1)

    ragrid = 5
    decgrid = 5
    rastep = 0.5
    decstep = 0.5

    ralo = floor(ralo/ragrid) *ragrid
    rahi = (1+floor(rahi/ragrid)) *ragrid
    declo = floor(declo/decgrid) *decgrid
    dechi = (1+floor(dechi/decgrid)) *decgrid

    for ra in arange(ralo,rahi,ragrid):
        xy=[]
        lastok = True
        for dec in arange(declo, dechi, decstep):
            (bhp, hpx, hpy) = radectohealpixf(ra, dec, nside)
            #print 'ra,dec', ra,dec, 'bighp', bhp, 'x,y', hpx,hpy
            if bhp != bighp:
                if lastok and len(xy):
                    plot([x for x,y in xy], [y for x,y in xy], 'r-')
                    xy = []
                lastok = False
                continue
            lastok = True
            xy.append((hpx,hpy))
        if len(xy):
            plot([x for x,y in xy], [y for x,y in xy], 'r-')

    for dec in arange(declo, dechi, decgrid):
        xy=[]
        lastok = True
        for ra in arange(ralo,rahi,rastep):
            (bhp, hpx, hpy) = radectohealpixf(ra, dec, nside)
            if bhp != bighp:
                if lastok and len(xy):
                    plot([x for x,y in xy], [y for x,y in xy], 'r-')
                    xy = []
                lastok = False
                continue
            lastok = True
            xy.append((hpx,hpy))
        if len(xy):
            plot([x for x,y in xy], [y for x,y in xy], 'r-')

    axhline(y=0, lw=0.5, ls='-', color='0.8')
    axhline(y=nside, lw=0.5, ls='-', color='0.8')
    axvline(x=0, lw=0.5, ls='-', color='0.8')
    axvline(x=nside, lw=0.5, ls='-', color='0.8')
    axis('scaled')
    axis([-0.1, nside+0.1, -0.1, nside+0.1])
    title('RA,Dec lines in healpix space, for big healpix %i' % bighp)
    savefig('hp-radec-%i.png' % bighp)

def plot_smallcircles_on_healpix(nside=1, bighp=1):
    clf()
    xlabel('Healpix x')
    ylabel('Healpix y')

    #xstep = ystep = 0.05
    #radius = 1.3

    xstep = ystep = 0.1
    radius = 2.5

    ncircle = 60

    runit = sqrt(deg2distsq(radius))

    figure(figsize=(6,6))
    for x in arange(0, 1.001, xstep):
        for y in arange(0, 1.001, ystep):
            xyz = array(healpix_to_xyz(bighp, nside, x, y))
            r,d = xyztoradec(xyz)
            dra,ddec = derivatives_at_radec(r,d)
            dra = dra[0,:]
            dra /= vector_norm(dra)
            ddec = ddec[0,:]
            ddec /= vector_norm(ddec)
            assert(abs(dot(xyz, dra))  < 1e-15)
            assert(abs(dot(xyz, ddec)) < 1e-15)
            assert(abs(dot(dra, ddec)) < 1e-15)
            #print dot(xyz,dra)
            #print dot(xyz,ddec)
            #print dot(dra,ddec)
            theta = linspace(0, 2*pi, ncircle)
            xyzs = array([xyz + runit*sin(t)*dra + runit*cos(t)*ddec
                          for t in theta])
            xyzs /= vector_norm(xyzs)
            #(bighp, hpx, hpy) = xyz_to_healpixf(p[0],p[1],p[2], nside)
            HPs = array([xyztohealpixf(p[0],p[1],p[2], nside) for p in xyzs])
            bighps = HPs[:,0]
            hpx = HPs[:,1]
            hpy = HPs[:,2]
            I = (bighps == bighp)
            if sum(I):
                plot(hpx[I], hpy[I], 'r-')

    axhline(y=0, lw=0.5, ls='-', color='0.8')
    axhline(y=nside, lw=0.5, ls='-', color='0.8')
    axvline(x=0, lw=0.5, ls='-', color='0.8')
    axvline(x=nside, lw=0.5, ls='-', color='0.8')
    axis('scaled')
    margin = 0.05
    axis([-margin, nside+margin, -margin, nside+margin])
    title('Small circles in healpix space, for big healpix %i' % bighp)
    xlabel('Healpix x')
    ylabel('Healpix y')
    savefig('hp-circles-%i.png' % bighp)

if __name__ == '__main__':

    for bighp in [1,5,9]:
        plot_smallcircles_on_healpix(nside=1, bighp=bighp)
        plot_radec_on_healpix(nside=1, bighp=bighp)

    plot_healpix_on_radec(nside=1, fn='healpix-1a.png')
    plot_healpix_on_radec(nside=3, fn='healpix-3a.png', textkws={'fontsize':8})

    plot_healpix_on_radec(nside=1, cosdec=True)
    plot_healpix_on_radec(nside=3, cosdec=True, textkws={'fontsize':8})





# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from numpy import *
import datetime
import numpy as np
from functools import reduce

arcsecperrad = 3600. * 180. / np.pi
axistilt = 23.44 # degrees

def ra_normalize(ra):
    return np.mod(ra, 360.)

def ra_ranges_overlap(ralo, rahi, ra1, ra2):
    ''' Given two ranges, [ralo,rahi], [ra1,ra2], returns True if they overlap.'''
    import numpy as np
    x1 = np.cos(np.deg2rad(ralo))
    y1 = np.sin(np.deg2rad(ralo))

    x2 = np.cos(np.deg2rad(rahi))
    y2 = np.sin(np.deg2rad(rahi))

    x3 = np.cos(np.deg2rad(ra1))
    y3 = np.sin(np.deg2rad(ra1))

    x4 = np.cos(np.deg2rad(ra2))
    y4 = np.sin(np.deg2rad(ra2))

    #cw31 = x1*y3 - x3*y1
    cw32 = x2*y3 - x3*y2

    cw41 = x1*y4 - x4*y1
    #cw42 = x2*y4 - x4*y2

    #print('3:', cw31, cw32)
    #print('4:', cw41, cw42)
    return np.logical_and(cw32 <= 0, cw41 >= 0)

#
def transform(long, lat, poleTo, poleFrom):
    (alphaGP,deltaGP) = deg2rad(poleFrom[0]), deg2rad(poleFrom[1])
    lCP = deg2rad(poleTo[0])
    alpha = deg2rad(long)
    delta = deg2rad(lat)
    ra = rad2deg(lCP - arctan2(sin(alpha - alphaGP),
                               tan(delta) * cos(deltaGP) - cos(alpha - alphaGP) * sin(deltaGP)))
    dec = rad2deg(arcsin((sin(deltaGP)*sin(delta) + cos(deltaGP)*cos(delta)*cos(alpha - alphaGP))))
    ra = ra_normalize(ra)
    return ra,dec

# Galactic (l,b) to equatorial (ra,dec).
# Lifted from LSST's afw.coord.Coord class by Steve Bickerton.
def lbtoradec(l, b):
    # aka 'GalacticPoleInFk5'
    poleTo   = (192.8595, 27.12825)
    # aka 'Fk5PoleInGalactic'
    poleFrom = (122.9320, 27.12825)
    return transform(l, b, poleTo, poleFrom)

galactictoradec = lbtoradec

def eclipticPoleInclination(epoch):
    T = (epoch - 2000.0) / 100.0
    eclincl = (23.0 + 26.0/60.0 +
               (21.448 - 46.82*T - 0.0006*T*T - 0.0018*T*T*T)/3600.0)
    return eclincl

# Thanks to Steve Bickerton in lsst.afw.Coord : EclipticCoord::toFk5
def ecliptictoradec(a, b, epoch=2000.):
    eclincl = eclipticPoleInclination(epoch)
    eclipticPoleInFk5 = (270.0, 90.0 - eclincl)
    fk5PoleInEcliptic = (90.0, 90.0 - eclincl)
    return transform(a, b, eclipticPoleInFk5, fk5PoleInEcliptic)

# Thanks to Steve Bickerton in lsst.afw.Coord : Fk5Coord::toEcliptic
def radectoecliptic(ra, dec, epoch=2000.):
    eclincl = eclipticPoleInclination(epoch)
    eclPoleInEquatorial = (270.0, 90.0 - eclincl)
    equPoleInEcliptic = (90.0, 90.0 - eclincl)
    return transform(ra, dec, equPoleInEcliptic, eclPoleInEquatorial)

# scalars (racenter, deccenter) in deg
# scalar radius in deg
# arrays (ra,dec) in deg
# returns array of booleans
def points_within_radius(racenter, deccenter, radius, ra, dec):
    return radecdotproducts(racenter, deccenter, ra, dec) >= cos(deg2rad(radius))

def points_within_radius_range(racenter, deccenter, radiuslo, radiushi, ra, dec):
    d = radecdotproducts(racenter, deccenter, ra, dec)
    return (d <= cos(deg2rad(radiuslo))) * (d >= cos(deg2rad(radiushi)))

# scalars (racenter, deccenter) in deg
# arrays (ra,dec) in deg
# returns array of cosines
def radecdotproducts(racenter, deccenter, ra, dec):
    xyzc = radectoxyz(racenter, deccenter).T
    xyz = radectoxyz(ra, dec)
    return dot(xyz, xyzc)[:,0]

# RA, Dec in degrees: scalars or 1-d arrays.
# returns xyz of shape (N,3)
def radectoxyz(ra_deg, dec_deg):
    ra  = deg2rad(ra_deg)
    dec = deg2rad(dec_deg)
    cosd = cos(dec)
    xyz = vstack((cosd * cos(ra),
                  cosd * sin(ra),
                  sin(dec))).T
    assert(xyz.shape[1] == 3)
    return xyz

# RA,Dec in degrees
# returns (dxyz_dra, dxyz_ddec)
def derivatives_at_radec(ra_deg, dec_deg):
    ra  = deg2rad(ra_deg)
    dec = deg2rad(dec_deg)
    cosd = cos(dec)
    sind = sin(dec)
    cosra = cos(ra)
    sinra = sin(ra)
    return (180./pi * vstack((cosd * -sinra,
                              cosd *  cosra,
                              0)).T,
            180./pi * vstack((-sind * cosra,
                              -sind * sinra,
                               cosd)).T)

def xyztoradec(xyz):
    '''
    Converts positions on the unit sphere to RA,Dec in degrees.

    'xyz' must be a numpy array, either of shape (3,) or (N,3)

    Returns a tuple (RA,Dec).

    If 'xyz' is a scalar, RA,Dec are scalars.

    If 'xyz' is shape (N,3), RA,Dec are shape (N,).

    >>> xyztoradec(array([1,0,0]))
    (0.0, 0.0)

    >>> xyztoradec(array([ [1,0,0], [0,1,0], [0,0,1]]))
    (array([  0.,  90.,   0.]), array([  0.,   0.,  90.]))

    >>> xyztoradec(array([0,1,0]))
    (90.0, 0.0)

    >>> xyztoradec(array([0,0,1]))
    (0.0, 90.0)

    '''
    if len(xyz.shape) == 1:
        # HACK!
        rs,ds = xyztoradec(xyz[newaxis,:])
        return (rs[0], ds[0])
    (nil,three) = xyz.shape
    assert(three == 3)
    ra = arctan2(xyz[:,1], xyz[:,0])
    ra += 2*pi * (ra < 0)
    dec = arcsin(xyz[:,2] / norm(xyz)[:,0])
    return (rad2deg(ra), rad2deg(dec))


#####################

# RA,Decs in degrees.  Both pairs can be arrays.
def distsq_between_radecs(ra1, dec1, ra2, dec2):
    xyz1 = radectoxyz(ra1, dec1)
    xyz2 = radectoxyz(ra2, dec2)
    # (n,3) (m,3)
    s0 = xyz1.shape[0]
    s1 = xyz2.shape[0]
    d2 = zeros((s0,s1))
    for s in range(s0):
        d2[s,:] = sum((xyz1[s,:] - xyz2)**2, axis=1)
    if s0 == 1 and s1 == 1:
        d2 = d2[0,0]
    elif s0 == 1:
        d2 = d2[0,:]
    elif s1 == 1:
        d2 = d2[:,0]
    return d2

# RA,Decs in degrees.
def distsq_between_radecs(ra1, dec1, ra2, dec2):
    '''
    Computes the distance-square on the unit sphere between two (arrays of) RA,Decs.
    '''
    xyz1 = radectoxyz(ra1, dec1)
    xyz2 = radectoxyz(ra2, dec2)
    # (n,3) (m,3)
    s0 = xyz1.shape[0]
    s1 = xyz2.shape[0]
    d2 = zeros((s0,s1))
    for s in range(s0):
        d2[s,:] = sum((xyz1[s,:] - xyz2)**2, axis=1)
    if s0 == 1 and s1 == 1:
        d2 = d2[0,0]
    elif s0 == 1:
        d2 = d2[0,:]
    elif s1 == 1:
        d2 = d2[:,0]
    return d2

# RA,Decs in degrees.
def arcsec_between(ra1, dec1, ra2, dec2):
    '''
    Computes the angle between two (arrays of) RA,Decs.

    >>> from numpy import round
    >>> print round(arcsec_between(0, 0, 1, 0), 6)
    3600.0

    >>> print round(arcsec_between(array([0, 1]), array([0, 0]), 1, 0), 6)
    [ 3600.     0.]

    >>> print round(arcsec_between(1, 0, array([0, 1]), array([0, 0])), 6)
    [ 3600.     0.]

    >>> print round(arcsec_between(array([0, 1]), array([0, 0]), array([0, 1]), array([0, 0])), 6)
    [[    0.  3600.]
     [ 3600.     0.]]

    '''
    return distsq2arcsec(distsq_between_radecs(ra1,dec1,ra2,dec2))

def degrees_between(ra1, dec1, ra2, dec2):
    return arcsec2deg(arcsec_between(ra1, dec1, ra2, dec2))

def deg2distsq(deg):
    return rad2distsq(deg2rad(deg))

def deg2dist(deg):
    return rad2dist(deg2rad(deg))

def rad2dist(r):
    return sqrt(rad2distsq(r))

def rad2distsq(r):
    # inverse of distsq2arc; cosine law.
    return 2.0 * (1.0 - cos(r));

def distsq2rad(dist2):
    return arccos(1. - dist2 / 2.)
def distsq2arcsec(dist2):
    return rad2arcsec(distsq2rad(dist2))
def distsq2deg(dist2):
    return rad2deg(distsq2rad(dist2))

def rad2deg(r):
    return 180.0*r/pi
def rad2arcsec(r):
    return 648000.0*r/pi

def arcsec2rad(a):
    return a*pi/648000.0
def arcsec2deg(a):
    return rad2deg(arcsec2rad(a))



# x can be an array of shape (N,D)
# returns an array of shape (N,1)
def norm(x):
    if len(x.shape) == 2:
        return sqrt(sum(x**2, axis=1))[:,newaxis]
    else:
        return sqrt(sum(x**2))
        
vector_norm = norm

# proper motion (dl, db, dra, or ddec) in mas/yr
# dist in kpc
# returns velocity in km/s
def pmdisttovelocity(pm, dist):
    # (pm in deg/yr) * (dist in kpc) to (velocity in km/s)
    pmfactor = 1/3.6e6 * pi/180. * 0.977813952e9
    return pm * dist * pmfactor


# ra, dec in degrees
# pmra = d(RA*cos(Dec))/dt, pmdec = dDec/dt, in deg/yr or mas/yr
# returns (l,b, pml,pmb) in degrees and [the same units as pmra,pmdec]
#    pml is d(l*cos(b))/dt
def pm_radectolb(ra, dec, pmra, pmdec):
    (l1, b1) = radectolb(ra, dec)
    # the Jo Bovy method:
    (a,d) = galactic_pole
    alphangp = deg2rad(a)
    deltangp = deg2rad(d)
    delta = deg2rad(dec)
    alpha = deg2rad(ra)
    b = deg2rad(b1)

    cosphi = ((sin(deltangp) - sin(delta)*sin(b)) /
              (cos(delta)*cos(b)))
    sinphi = ((sin(alpha - alphangp) * cos(deltangp)) /
              cos(b))

    dlcosb =  cosphi * pmra + sinphi * pmdec
    db     = -sinphi * pmra + cosphi * pmdec
    return (l1, b1, dlcosb, db)

# ra, dec in degrees
# returns (l,b) in degrees
def radectolb(ra, dec):
    (xhat, yhat, zhat) = galactic_unit_vectors()
    xyz = radectoxyz(ra, dec)
    xg = dot(xyz, xhat)
    yg = dot(xyz, yhat)
    zg = dot(xyz, zhat)
    # danger, will robinson, danger!
    # abuse the xyztoradec routine to convert xyz in the galactic
    # unit sphere to (l,b) in galactic coords.
    (l,b) = xyztoradec(hstack((xg, yg, zg)))
    # galactic system is left-handed so "l" comes out backward.
    l = 360. - l
    return (l,b)

# ra,dec in degrees
# dist in kpc
# pmra is d(ra * cos(dec))/dt  in mas/yr
# pmdec is in mas/yr
# returns (pmra, pmdec) in the same units
def remove_solar_motion(ra, dec, dist, pmra, pmdec):
    (xhat, yhat, zhat) = galactic_unit_vectors()
    # (we only need yhat)
    # V_sun in kpc / yr
    vsun = 240. * 1.02268944e-9 * yhat.T
    # unit vectors on celestial sphere
    unitxyz = radectoxyz(ra, dec)
    # heliocentric positions in kpc
    xyz = dist[:,newaxis] * unitxyz
    # numerical difference time span in yr
    dyr = 1.
    # transverse displacements on celestial unit sphere
    unitxyz2 = radectoxyz(ra  + pmra/cos(deg2rad(dec)) /3.6e6 * dyr,
                          dec + pmdec/3.6e6 * dyr)
    # heliocentric transverse displacement of the observed star in kpc
    dxyz = (unitxyz2 - unitxyz) * dist[:,newaxis]
    # galactocentric displacement in kpc
    dxyz -= vsun * dyr
    # new 3-space position in kpc
    xyz3 = xyz + dxyz
    # back to the lab, deg
    (ra3,dec3) = xyztoradec(xyz3)
    # adjusted angular displacement, deg
    dra = ra3 - ra
    # tedious RA wrapping
    dra += 360. * (dra < -180)
    dra -= 360. * (dra >  180)
    # convert back to proper motions
    return ((dra * cos(deg2rad(dec3)) / dyr) * 3.6e6,
            ((dec3 - dec) / dyr) * 3.6e6)


def axis_angle_rotation_matrix(axis, angle):
    '''
    axis: 3-vector about which to rotate
    angle: angle about which to rotate, in degrees.

    Returns: 3x3 rotation matrix
    '''
    theta = np.deg2rad(angle)
    ct = np.cos(theta)
    st = np.sin(theta)
    ux,uy,uz = axis / np.sqrt(np.sum(axis**2))
    R = np.array([
        [ct + ux**2*(1-ct),
         ux*uy*(1-ct) - uz*st,
         ux*uz*(1-ct) + uy*st],
        [uy*ux*(1-ct) + uz*st,
          ct + uy**2*(1-ct),
          uy*uz*(1-ct)-ux*st],
        [uz*ux*(1-ct)-uy*st,
         uz*uy*(1-ct)+ux*st,
         ct+uz**2*(1-ct)],
    ])
    return R


# the north galactic pole, (RA,Dec), in degrees, from Bovy.
# This matches Schlegel's email of 2015-02-19 citing the
# Hipparcos explanatory supplement.
galactic_pole = (192.85948, 27.12825)
# vs Wikipedia's (192.859508, 27.128336)

# This puts (RA,DEC) = (1,1) at (l,b) = (98.941031, -59.643798).

# returns (xhat, yhat, zhat), unit vectors in the RA,Dec unit sphere
# of the galactic coordinates.
def galactic_unit_vectors():
    # Galactic longitude of celestial equator
    lomega = 32.93192
    # direction to Galactic Pole
    zhat = radectoxyz(*galactic_pole).T

    # where the galactic plane crosses the equatorial plane
    X = np.cross(zhat.T, np.array([[0,0,-1],]))
    X /= np.sqrt(np.sum(X**2))

    # Rotate X by lomega around zhat.
    Rx = axis_angle_rotation_matrix(zhat[:,0], -lomega)
    Ry = axis_angle_rotation_matrix(zhat[:,0], 90.-lomega)

    xhat = np.dot(Rx, X.T)
    yhat = -np.cross(zhat.T, xhat.T).T

    return (xhat, yhat, zhat)


def mjdtodate(mjd):
    jd = mjdtojd(mjd)
    return jdtodate(jd)

def jdtodate(jd):
    unixtime = (jd - 2440587.5) * 86400. # in seconds
    return datetime.datetime.utcfromtimestamp(unixtime)

def mjdtojd(mjd):
    return mjd + 2400000.5

def jdtomjd(jd):
    return jd - 2400000.5

def timedeltatodays(dt):
    return dt.days + (dt.seconds + dt.microseconds/1e6)/86400.

def datetomjd(d):
    d0 = datetime.datetime(1858, 11, 17, 0, 0, 0)
    dt = d - d0
    # dt is a timedelta object.
    return timedeltatodays(dt)

def datetojd(d):
    return mjdtojd(datetomjd(d))

# UTC for 2000 January 1.5
J2000 = datetime.datetime(2000,1,1,12,0,0,0,tzinfo=None)
# -> jd 2451545.0

def ecliptic_basis(eclipticangle = 23.439281):
    Equinox= array([1,0,0])
    CelestialPole = array([0,0,1])
    YPole = cross(CelestialPole, Equinox)
    EclipticAngle= deg2rad(eclipticangle)
    EclipticPole= (CelestialPole * cos(EclipticAngle) - YPole * sin(EclipticAngle))
    Ydir = cross(EclipticPole, Equinox)
    return (Equinox, Ydir, EclipticPole)

meters_per_au = 1.4959e11 # thanks, Google
speed_of_light = 2.99792458e8 # m/s
seconds_per_day = 86400.
days_per_year = 365.25

def days_to_years(d):
    return d / days_per_year

def au_to_meters(au):
    return au * meters_per_au

def seconds_to_days(s):
    return s / seconds_per_day

# Returns the light travel time for the given distance (in AU), in days.
def au_light_travel_time_days(au):
    return seconds_to_days(au_to_meters(au) / speed_of_light)
    


def hms2ra(h, m, s):
    return 15. * (h + (m + s/60.)/60.)

def tokenize_hms(s):
    s = s.strip()
    tokens = s.split()
    tokens = reduce(list.__add__, [t.split(':') for t in tokens])
    h = len(tokens) >= 1 and float(tokens[0]) or 0
    m = len(tokens) >= 2 and float(tokens[1]) or 0
    s = len(tokens) >= 3 and float(tokens[2]) or 0
    return (h,m,s)

def hmsstring2ra(st):
    '''
    >>> st = "00 44 02.08"

    >>> hmsstring2ra(st)
    11.008666666666667

    >>> ra2hmsstring(hmsstring2ra(st), sec_digits=2) == st
    True
    
    '''
    (h,m,s) = tokenize_hms(st)
    return hms2ra(h, m, s)

def dms2dec(sign, d, m, s):
    return sign * (d + (m + s/60.)/60.)

def dmsstring2dec(s):
    sign = (s[0] == '-') and -1.0 or 1.0
    if s[0] == '-' or s[0] == '+':
        s = s[1:]
    (d,m,s) = tokenize_hms(s)
    return dms2dec(sign, d, m, s)

# RA in degrees
def ra2hms(ra):
    ra = ra_normalize(ra)
    h = ra * 24. / 360.
    hh = int(floor(h))
    m = (h - hh) * 60.
    mm = int(floor(m))
    s = (m - mm) * 60.
    return (hh, mm, s)

# Dec in degrees
def dec2dms(dec):
    sgn = (dec >= 0) and 1. or -1.
    d = dec * sgn
    dd = int(floor(d))
    m = (d - dd) * 60.
    mm = int(floor(m))
    s = (m - mm) * 60.
    if s >= 60.:
        m += 1.
        s -= 60.
    # don't just return sgn*d because values between 0 and 1 deg will get you!
    return (sgn, d, m, s)

# RA in degrees
def ra2hmsstring(ra, separator=' ', sec_digits=3):
    (h,m,s) = ra2hms(ra)
    #print 'hms', h,m,s
    ss = int(floor(s))
    #ds = int(round((s - ss) * 1000.0))
    # fractional seconds
    fs = s - ss
    #print 'ss,fs', ss, fs
    fracstr = '%.*f' % (sec_digits, fs)
    #print 'fracstr', fracstr

    if fs >= 1.:
        ss += 1
        fs -= 1.

    if sec_digits > 0:
        fracstr = '%.*f' % (sec_digits, fs)
        if fracstr[0] == '1':
            ss += 1
            fs -= 1.
    if ss >= 60:
        ss -= 60
        m += 1
    if m >= 60:
        m -= 60
        h += 1
    if sec_digits == 0:
        sstr = '%0.2i' % (ss)
    else:
        #sfmt = '%%0.2i.%%0.%ii' % (sec_digits)
        #sstr = sfmt % (ss, ds)
        sstr = '%0.2i' % ss
        # fractional seconds string -- 0.XXX
        fracstr = '%.*f' % (sec_digits, fs)
        #print 'fracstr', fracstr
        if fracstr[0] == '-':
            fracstr = fracstr[1:]
        assert(fracstr[0] == '0')
        sstr += fracstr[1:]
    return separator.join(['%0.2i' % h, '%0.2i' % m, sstr])

# Dec in degrees
def dec2dmsstring(dec, separator=' ', sec_digits=3):
    '''
    >>> dec2dmsstring(41.5955538864, sec_digits=3)
    '+41 35 43.994'

    >>> dec2dmsstring(41.5955538864, sec_digits=2)
    '+41 35 43.99'

    >>> dec2dmsstring(41.5955538864, sec_digits=1)
    '+41 35 44.0'

    '''
    (sgn, d,m,s) = dec2dms(dec)
    ss = int(floor(s))
    fs = s - ss
    if sgn > 0:
        signc = '+'
    else:
        signc = '-'
    if sec_digits == 0:
        sstr = '%0.2i' % (ss)
    else:
        # fractional seconds string -- 0.XXX
        fracstr = '%.*f' % (sec_digits, fs)
        # but it can be 1.00 ...
        #print 'dec fracstr', fracstr
        if fracstr[0] == '1':
            ss += 1
        if ss >= 60:
            ss -= 60
            m += 1
            if m >= 60:
                m -= 60
                d += 1
        sstr = '%0.2i' % ss + fracstr[1:]

    return separator.join(['%c%0.2i' % (signc, d), '%0.2i' % m, sstr])

def xyzarrtoradec(xyz):
    return (degrees(xy2ra(xyz[0], xyz[1])), degrees(z2dec(xyz[2])))

def deg2rad(d):    return d*pi/180.0
def deg2arcmin(d): return d * 60.
def deg2arcsec(d): return d * 3600.
def rad2arcmin(r): return 10800.0*r/pi
def arcmin2rad(a): return a*pi/10800.0
def arcmin2deg(a): return a/60.
def arcmin2rad(a): return deg2rad(arcmin2deg(a))
def radec2x(r,d):  return cos(d)*cos(r) # r,d in radians
def radec2y(r,d):  return cos(d)*sin(r) # r,d in radians
def radec2z(r,d):  return sin(d)        # r,d in radians
def z2dec(z):      return asin(z)     # result in radians
def xy2ra(x,y):
    "Convert x,y to ra in radians"
    r = atan2(y,x)
    r += 2*pi*(r<0.)
    return r


def rad2distsq(rad):
    return 2. * (1. - cos(rad))

def arcsec2distsq(arcsec):
    return rad2distsq(arcsec2rad(arcsec))

def arcsec2dist(arcsec):
    return sqrt(arcsec2distsq(arcsec))

def arcmin2distsq(arcmin):
    return rad2distsq(arcmin2rad(arcmin))

def arcmin2dist(arcmin):
    return sqrt(arcmin2distsq(arcmin))

def dist2arcsec(dist):
    return distsq2arcsec(dist**2)

def dist2deg(dist):
    return distsq2deg(dist**2)




if __name__ == '__main__':
    import doctest
    doctest.testmod()

    assert(ra_ranges_overlap(359, 1, 0.5, 1.5) == True)
    assert(ra_ranges_overlap(359, 1, 358, 0.)  == True)
    assert(ra_ranges_overlap(359, 1, 358, 2.)  == True)
    assert(ra_ranges_overlap(359, 1, 359.5, 0.5) == True)
    assert(ra_ranges_overlap(359, 1, 357, 358) == False)
    assert(ra_ranges_overlap(359, 1, 2, 3) == False)
    assert(ra_ranges_overlap(359, 1, 179, 181) == False)
    assert(ra_ranges_overlap(359, 1, 90, 270) == False)



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import sys

#from numpy import array, matrix, linalg
from numpy import *
from numpy.random import *
from numpy.linalg import *
from matplotlib.pylab import figure, plot, xlabel, ylabel, loglog, clf
from matplotlib.pylab import semilogy
#from pylab import *

class Transform(object):
    scale = None
    rotation = None
    incenter = None
    outcenter = None

    def apply(self, X):
        #print X
        dx = X - self.incenter
        #print dx
        dx = dx * self.scale
        #print dx
        dx = self.rotation * dx
        #print dx
        dx = dx + self.outcenter
        #print dx
        return dx

    def __str__(self):
        s = ('<Transform: tin (%f,%f) scale (%f) rot (%f, %f; %f, %f) tout (%f, %f)>' %
             (self.incenter[0], self.incenter[1], self.scale,
              self.rotation[0,0], self.rotation[0,1], self.rotation[1,0], self.rotation[1,1],
              self.outcenter[0], self.outcenter[1]))
        return s

def procrustes(X, Y):
    T = Transform()
    sx = X.shape
    if sx[0] != 2:
        print('X must be 2xN')
    sy = Y.shape
    if sy[0] != 2:
        print('Y must be 2xN')
    N = sx[1]

    mx = X.mean(axis=1).reshape(2,1)
    my = Y.mean(axis=1).reshape(2,1)
    #print 'mean(X) is\n', mx
    #print 'mean(Y) is\n', my
    T.incenter = mx
    T.outcenter = my

    #print 'X-mx is\n', X-mx
    #print '(X-mx)^2 is\n', (X-mx)*(X-mx)
    varx = sum(sum((X - mx)*(X - mx)), axis=1)
    vary = sum(sum((Y - my)*(Y - my)), axis=1)
    #print 'var(X) is', varx
    #print 'var(Y) is', vary
    T.scale = sqrt(vary / varx)
    #print 'scale is', T.scale

    C = zeros((2,2))
    for i in [0,1]:
        for j in [0,1]:
            C[i,j] = sum((X[i,:] - mx[i]) * (Y[j,:] - my[j]))
    #print 'cov is\n', C

    U,S,V = svd(C)
    U = matrix(U)
    V = matrix(V)
    
    #print 'U is\n', U
    #print 'U\' is\n', U.transpose()
    #print 'V is\n', V
    R = V * U.transpose()
    #print 'R is\n', R
    T.rotation = R
    return T


def test_procrustes_1():
    # Create a Transform, apply it to some points, then run procrustes to see if we
    # recover the Transform exactly.
    t1 = Transform()
    t1.scale = 3.0
    A = 48.0 * pi/180.0
    t1.rotation = matrix([[sin(A), cos(A)], [-cos(A), sin(A)]])
    t1.incenter = array([42, 500]).reshape(2,1)
    t1.outcenter = array([600, -12]).reshape(2,1)

    N = 4
    pts = zeros((2,N))
    tpts = zeros((2,N))
    for i in range(N):
        pts[0,i] = t1.incenter[0] + ((i % 2) - 0.5) * 200
        pts[1,i] = t1.incenter[1] + (((i/2) % 2) - 0.5) * 200

    for i in range(N):
        pt = pts[:,i].reshape(2,1)
        tpts[:,i] = t1.apply(pt).reshape(1,2)

    t2 = procrustes(pts, tpts)

    print('pts:', pts)
    print('tpts:', tpts)

    print('t1 is', t1)
    print('t2 is', t2)


def draw_sample(inoise=1, fnoise=0, iqnoise=-1,
                dimquads=4, quadscale=100, imgsize=1000,
                Rsteps=10, Asteps=36):

    # Stars that compose the field quad.
    fquad = zeros((2,dimquads))
    fquad[0,0] = imgsize/2 - quadscale/2
    fquad[1,0] = imgsize/2
    fquad[0,1] = imgsize/2 + quadscale/2
    fquad[1,1] = imgsize/2
    for i in range(2, dimquads):
        fquad[0,i] = imgsize/2 + randn(1) * quadscale
        fquad[1,i] = imgsize/2 + randn(1) * quadscale

    # Index quad is field quad plus jitter.
    iquad = fquad + randn(*fquad.shape)

    # Solve for transformation
    T = procrustes(iquad, fquad)

    # Put the index quad stars through the transformation
    itrans = zeros(fquad.shape)
    for i in range(dimquads):
        fq = fquad[:,i].reshape(2,1)
        itrans[:,i] = T.apply(fq).transpose()

    # Field quad center...
    qc = mean(fquad, axis=1)

    # Sample stars on a R^2, theta grid.
    #rads = sqrt((array(range(Rsteps))+1) / float(Rsteps)) * imgsize/2
    N = Rsteps * Asteps
    rads = sqrt((array(range(Rsteps))+0.5) / float(Rsteps)) * imgsize/2
    thetas = array(range(Asteps)) / float(Asteps) * 2.0 * pi
    fstars = zeros((2,N))
    for r in range(Rsteps):
        for a in range(Asteps):
            fstars[0, r*Asteps + a] = sin(thetas[a]) * rads[r] + qc[0]
            fstars[1, r*Asteps + a] = cos(thetas[a]) * rads[r] + qc[1]
    # Put them through the transformation...
    istars = zeros((2,N))
    for i in range(N):
        fs = fstars[:,i].reshape(2,1)
        istars[:,i] = T.apply(fs).transpose()

    R = sqrt((fstars[0,:] - qc[0])**2 + (fstars[1,:] - qc[1])**2)
    E = sqrt(sum((fstars - istars)**2, axis=0))

    # Fit to a linear model...
    xfit = R**2
    yfit = E**2
    A = zeros((2,N))
    A[0,:] = 1
    A[1,:] = xfit.transpose()
    (C,resids,rank,s) = lstsq(A.transpose(), yfit)

    return (fquad, iquad, T, itrans, qc, fstars, istars,
            R, E, C)

if __name__ == '__main__':

    test_procrustes_1()
    sys.exit(0)

    #N = 1000
    N = 100
    C = zeros((2,N))
    QD = zeros((N))
    for i in range(N):
        (fquad, iquad, T, itrans, qc, fstars, istars, R, E, c) = draw_sample()
        C[:,i] = c
        QD[i] = sqrt(sum((iquad - fquad)**2) / 4.0)

    C0 = C[0,:]
    C1 = C[1,:]

    figure(1)
    clf()
    loglog(C0, C1, 'b.')
    xlabel('E^2 vs R^2 - Fit coefficient 0')
    ylabel('E^2 vs R^2 - Fit coefficient 1')

    figure(2)
    clf()
    semilogy(QD, C1, 'b.')
    xlabel('Field-to-Index Quad Mean Distance')
    ylabel('E^2-vs-R^2 fit linear coefficient')

    #semilogy(QD, C1, 'bo')
    #xlabel('Quad Distance')
    #ylabel('C1')


    #figure(1)
    #I=[0,2,1,3,0];
    #plot(fquad[0,I], fquad[1,I], 'bo-', itrans[0,I], itrans[1,I], 'ro-')

    #figure(2)
    #plot(fstars[0,:], fstars[1,:], 'b.', istars[0,:], istars[1,:], 'r.')

    #figure(1)
    #I=[0,2,1,3,0];
    #plot(fquad[0,I], fquad[1,I], 'bo-',
    #     itrans[0,I], itrans[1,I], 'ro-',
    #     fstars[0,:], fstars[1,:], 'b.',
    #     istars[0,:], istars[1,:], 'r.')

    #figure(2)
    #plot(R, E, 'r.')
    #xlabel('R')
    #ylabel('E')

    #figure(3)
    #plot(R**2, E**2, 'r.')
    #xlabel('R^2')
    #ylabel('E^2')

    #print 'Fit coefficients are', C

    #figure(2)
    #xplot = array(range(101)) / 100.0 * max(xfit)
    #plot(R**2, E**2, 'r.',
    #     xplot, C[0] + C[1]*xplot, 'b-')
    #xlabel('R^2')
    #ylabel('E^2')

    #show()




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from astrometry.util.fits import *
from numpy import logical_and

import optparse
import sys

if __name__ == '__main__':
    parser = optparse.OptionParser(usage='%prog [options] <input filename (2mass_hpXXX.fits)> <output-filename>')
    parser.add_option('-b', dest='band', help='Select the band on which to apply cuts: "J" (default), "H", or "K"')
    parser.set_defaults(band='J')

    opt,args = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        sys.exit(-1)

    lband = opt.band.lower()
    infn = args[0]
    outfn = args[1]

    print('Reading %s, writing %s' % (infn, outfn))
    T = fits_table(infn)
    qual_col = '%s_quality' % lband
    cc_col = '%s_cc' % lband
    mag_col = '%s_mag' % lband
    qual = T.getcolumn(qual_col)
    cc = T.getcolumn(cc_col)

    # if ((entry->j_quality != TWOMASS_QUALITY_NO_BRIGHTNESS) &&
    #     (entry->j_cc == TWOMASS_CC_NONE)) {
        
    I = logical_and(qual != chr(0), cc == chr(0))
    print('Keeping %i of %i sources' % (sum(I), len(I)))

    T[I].write_to(outfn, columns=['ra','dec',mag_col])





# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import re
import datetime
from astrometry.util.starutil_numpy import *
import numpy as np

## FIXME -- requires at least one digit before the decimal place.
floatre = r'[+-]?[\d]+(.[\d]*)?([eE][+-]?[\d]*)?'

def floatvar(x):
    return '(?P<%s>%s)' % (x, floatre)

# For parsing orbital elements type output.
elemrexstr = (r'^' + floatvar('jd')
              + '.*' + ' EC= ' + floatvar('e')
              + '.*' + ' IN= ' + floatvar('i')
              + '.*' + ' OM= ' + floatvar('Omega')
              + ' W = ' + floatvar('pomega')
              + '.*' + 'MA= ' + floatvar('M')
              + '.*' + 'A = ' + floatvar('a')
              )
elemrex = re.compile(elemrexstr, re.MULTILINE | re.DOTALL)

# For finding "System GM" -- only in orbital elements type
sysgmrexstr = '^System GM *: ' + floatvar('gm') + r' '#AU\^3/d\^2$'
sysgmrex = re.compile(sysgmrexstr)

# For parsing X,V type output
xvrexstr = ('^' + floatvar('jd') + r' = (?P<ad>A\.D\. .*?)'
            + r'^\s+' + floatvar('x0')
            + ' +' + floatvar('x1')
            + ' +' + floatvar('x2')
            + '.*?'
            + r'^\s+' + floatvar('v0')
            + ' +' + floatvar('v1')
            + ' +' + floatvar('v2')
            )
xvrex = re.compile(xvrexstr, re.MULTILINE | re.DOTALL)

# For parsing "observer" type output (RA,Dec)
# With QUANTITIES=1; angle format=DEG
radecrexstr = (
    '^ ' + '(?P<datetime>[\d]{4}-[\w]{3}-[\d]{2} [\d]{2}:[\d]{2}) .*?'
               + floatvar('ra') + ' *?' + floatvar('dec'))
radecrex = re.compile(radecrexstr, re.MULTILINE | re.DOTALL)

'''
For output like:

#Date       UT      R.A. (J2000) Decl.    Delta     r     El.    Ph.   m1     Sky Motion
#            h m s                                                            "/min    P.A.
# geocentric
2007 04 01 000000   23.1983  -02.463     2.953   2.070   22.9  10.8  17.1    1.45    057.9
'''
#radecrexstr2 = (
#    '^' + '(?P<datetime>[\d]{4} [\d]{2} [\d]{2} [\d]{2}:[\d]{2}) .*?'
#               + floatvar('ra') + ' *?' + floatvar('dec'))
#radecrex = re.compile(radecrexstr, re.MULTILINE | re.DOTALL)



# Returns a list of lists of elements, plus a list of the JDs.
#   ([jd1, jd2, ...], [   [a1, e1, i1, Omega1, pomega1, M1, GM1], ... ])
# Where  i, Omega, pomega, M   are in radians
def parse_orbital_elements(s, needSystemGM=True):
    if needSystemGM:
        m = sysgmrex.search(s)
        if not m:
            print('Did not find "System GM" entry')
            return None
        gm = float(m.group('gm'))
    else:
        gm = 1.
    allE = []
    alljd = []
    for m in elemrex.finditer(s):
        d = m.groupdict()
        E = [np.deg2rad(x) if rad else x
             for (x,rad) in zip([float(d[x]) for x in
                                 ['a',  'e',   'i',  'Omega', 'pomega','M' ]],
                                [False, False, True, True,    True,    True])]
        E.append(gm)
        allE.append(E)
        alljd.append(float(d['jd']))
    return alljd,allE

# Returns (x, v, jd), each as numpy arrays.
#     x in AU
#     v in AU/day
def parse_phase_space(s):
    all_x = []
    all_v = []
    all_jd = []
    for m in xvrex.finditer(s):
        d = m.groupdict()
        x = np.array([float(d[k]) for k in ['x0','x1','x2']])
        v = np.array([float(d[k]) for k in ['v0','v1','v2']])
        all_x.append(x)
        all_v.append(v)
        all_jd.append(float(d['jd']))
    return (np.array(all_x), np.array(all_v), np.array(all_jd))

# Returns (ra,dec,jd), each as numpy arrays.
#   RA,Dec in J2000 deg
def parse_radec(s):
    all_ra = []
    all_dec = []
    all_jd = []
    for m in radecrex.finditer(s):
        d = m.groupdict()
        all_ra.append(float(d['ra']))
        all_dec.append(float(d['dec']))
        t = datetime.datetime.strptime(d['datetime'], '%Y-%b-%d %H:%M')
        # 2000-Jan-01 12:00
        all_jd.append(datetojd(t))
    return (np.array(all_ra), np.array(all_dec), np.array(all_jd))



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function

'''
This code allows one to interact with the JPL Horizons ephemeris
service via their telnet interface.  Yes, telnet.
'''

# ohhhhhh yeahhhhh
import telnetlib as tn

import datetime

import time

#from bmjd import Eph

from astrometry.util.file import *
from astrometry.util.starutil_numpy import *
from astrometry.util.fits import *
import sys


'''
earthfromssb-daily.eph:

Ephemeris Type [change] :  VECTORS
Target Body [change] :  Earth [Geocenter] [399]
Coordinate Origin [change] :  Solar System Barycenter (SSB) [500@0]
Time Span [change] :  Start=2003-04-28, Stop=2011-04-26, Step=1 d
Table Settings [change] :  CSV format=YES
Display/Output [change] :  plain text
'''

class Eph(object):
    def __init__(self, fn='earthfromssb-daily.eph', txt=None, linedelim='\n'):
        #self.entries = []
        if txt is not None:
            lines = txt.split(linedelim)
        else:
            lines = read_file(fn).split('\n')
        start = lines.index('$$SOE')
        end = lines.index('$$EOE')
        lines = lines[start+1:end]
        xyz = []
        jds = []
        lts = []
        for line in lines:
            X = line.split(',')
            X = [x.strip() for x in X]
            X = [x for x in X if len(x)]
            jd,ct,x,y,z,vx,vy,vz,lt,rg,rr = X
            x = float(x)
            y = float(y)
            z = float(z)
            jd = float(jd)
            lt = float(lt)
            rg = float(rg)
            lt *= 86400 # days to seconds
            #self.entries.append(jd, ct, x,y,z, lt, rg)
            xyz.append((x,y,z))
            jds.append(jd)
            lts.append(lt)
        self.entries = tabledata()
        self.entries.xyz = np.array(xyz)
        self.entries.jd = np.array(jds)
        self.entries.lt = np.array(lts)

    def get_entries_bounding_jd(self, jd):
        i = self.entries.jd.searchsorted(jd) - 1
        assert(i > 0)
        assert(i < (len(self.entries)-1))
        return self.entries[i], self.entries[i+1]



# Negotiate telnet options.  These ones seemed to be necessary to satisfy JPL c. 2012-02-27
class optcallback(object):
    def __init__(self, debug=False):
        self.debug = debug
    def __call__(self, socket, command, option):
        if self.debug: print('optcallback: socket', socket, 'command #', ord(command), 'option #', ord(option))
        cnum = command
        onum = option
        if cnum == tn.WILL and onum in [ tn.ECHO, tn.SGA ]:
            if self.debug: print('Got WILL', onum)
            if self.debug: print('Sending IAC DO', onum)
            socket.send(tn.IAC + tn.DO + onum)
        elif cnum == tn.WILL:
            if self.debug: print('Got WILL', onum)
            if self.debug: print('Sending IAC DONT', onum)
            socket.send(tn.IAC + tn.DONT + onum)
        elif cnum == tn.DO and onum == tn.NAWS:
            if self.debug: print('Got DO NAWS')
            # 16-bit big-endian W x H (0x00, 0x80) = 128
            reply = (tn.IAC + tn.WILL + tn.NAWS +
                     tn.IAC + tn.SB + tn.NAWS +
                     chr(0) + chr(0x80) + chr(0) + chr(0x80) +
                     tn.IAC + tn.SE)
            if self.debug:
                print('Replying:', reply)
                print(' hex: ', end=' ')
                for c in reply:
                    print('0x%02x' % ord(c), end=' ')
                print()
            socket.send(reply)
        elif cnum == tn.DO and onum == tn.TTYPE:
            if self.debug: print('Got DO TTYPE')
            reply = (tn.IAC + tn.WILL + tn.TTYPE +
                     tn.IAC + tn.SB + tn.TTYPE + chr(0) +
                     'DEC-VT100' +
                     tn.IAC + tn.SE)
            if self.debug:
                print('Replying:', reply)
                print(' hex: ', end=' ')
                for c in reply:
                    print('0x%02x' % ord(c), end=' ')
                print()
            socket.send(reply)
        elif cnum == tn.DO: # and onum == tn.TTYPE:
            if self.debug: print('Got DO', onum)
            if self.debug: print('Sending WONT', onum)
            socket.send(tn.IAC + tn.WONT + onum) #tn.TTYPE)

        
'''
About VT100-speak:

http://www.termsys.demon.co.uk/vtansi.htm
\x1b
7
\x1b
[r
\x1b
[999;999H
                <ESC>[{ROW};{COLUMN}HCursor Home 
\x1b
[6n

Query Cursor Position<ESC>[6n
    Requests a Report Cursor Position response from the device.
    
Report Cursor Position<ESC>[{ROW};{COLUMN}R
    Generated by the device in response to a Query Cursor Position request; reports current cursor position.
        
'''

'''
Telnet options are listed here:
  http://www.iana.org/assignments/telnet-options

Telnet(horizons.jpl.nasa.gov,6775): recv '\xff\xfb\x01'
Telnet(horizons.jpl.nasa.gov,6775): IAC WILL 1

(1: echo)

Telnet(horizons.jpl.nasa.gov,6775): recv '\xff\xfb\x03\xff\xfd\x1f\xff\xfd\x18'
Telnet(horizons.jpl.nasa.gov,6775): IAC WILL 3
Telnet(horizons.jpl.nasa.gov,6775): IAC DO 31
Telnet(horizons.jpl.nasa.gov,6775): IAC DO 24

(3: suppress go-ahead)
(31: negotiate about window size)
(24: terminal type)

    IAC SB NAWS <16-bit value> <16-bit value> IAC SE

    Sent by the Telnet client to inform the Telnet server of the
    window width and height.

    IAC SB TERMINAL-TYPE IS ... IAC SE

'''

def _horizons_login(debug=False):
    t = tn.Telnet('horizons.jpl.nasa.gov', 6775)
    print('Waiting for Horizons...')
    if debug:
        t.set_debuglevel(10)
    '''
    Each time a telnet option is read on the input flow, this callback
    (if set) is called with the following parameters : callback(telnet
    socket, command (DO/DONT/WILL/WONT), option). No other action is
    done afterwards by telnetlib.
    '''
    cb = optcallback(debug=debug)
    t.set_option_negotiation_callback(cb)

    # Horizons used to assume VT100, we think; it used to do:
    # # VT100: how big is your terminal?   '\x1b7\x1b[r\x1b[999;999H\x1b[6n'
    # ESC = chr(0x1b)
    # txt = t.read_until(ESC + '[6n', 30)
    # # big enough!
    # t.write(ESC + '[50;150R')

    if debug: print('Waiting for Horizons prompt')
    txt = t.read_until('Horizons>', 30)
    # Don't do page breaks
    t.write('PAGE\n')
    if debug: print('Waiting for Horizons prompt')
    txt = t.read_until('Horizons>', 30)
    return t

def get_radec_for_jds(bodyname, jd0, jd1, interval='1d', debug=False):
    t = _horizons_login(debug=debug)
    t.write( # Body name; if found, it asks "Continue?"
             '%s\r\n' % bodyname)
    #t.read_until('[A]pproaches, [E]phemeris, [F]tp,')

    #time.sleep(5)
    #txt = t.read_eager()
    #print 'Read', txt
    #time.sleep(1)
    #txt = t.read_eager()
    #print 'Read', txt

    #txt2 = t.read_until('Horizons> ')
    #txt2 = t.read_until('<cr>: ')
    txt=''
    txt2 = t.read_until('<cr>')
    print()
    print('--------------------------------')
    print()
    print(txt, txt2)
    print()
    print('--------------------------------')
    print()
    t.write('\n')

    # if 'EXACT' in txt2:
    #     '''
    #     >EXACT< designation search [CASE & SPACE sensitive]:
    #     DES = C/2014 Q2;
    #     Continue [ <cr>
    #     Telnet(horizons.jpl.nasa.gov,6775): send 'E\r\n'
    #     Telnet(horizons.jpl.nasa.gov,6775): recv ' n=no, ? ] : '
    #     Telnet(horizons.jpl.nasa.gov,6775): recv 'E\r\nContinue [ <cr>=yes, n=no, ? ] : '
    #     '''
    #     #txt = t.read_until('Continue')
    #     txt = t.read_until('Continue [ <cr>=yes,  n=no, ? ] :')
    #     print 'Read', txt
    #     t.write('\n')
        
    # t.write('\n')
    # txt = t.read_until('\nHorizons> ')
    # print txt
    txt = t.read_until('[E]phemeris')
    t.write('E\r\n')
    
    txt = t.read_until('Observe, Elements, Vectors  [o,e,v,?]')
    print(txt)
    t.write('o\n')
    t.read_until('Coordinate center [ <id>,coord,geo  ]')
    t.write('geo\n')
    t.read_until('Starting UT')
    t.write('JD %.9f\n' % jd0)
    t.read_until('Ending   UT')
    t.write('JD %.9f\n' % jd1)
    t.read_until('Output interval [ex: 10m, 1h, 1d, ? ]')
    t.write('%s\n' % interval)
    t.read_until('Accept default output [ cr=(y), n, ?]')
    t.write('\n')
    t.read_until('Select table quantities [ <#,#..>, ?]')
    t.write('1\n')

    # t.write('n\n')
    # t.read_until('Select table quantities')
    # t.write('1\n')
    # t.read_until('Output reference frame')
    # t.write('J2000\n')
    # t.read_until('Time-zone correction')
    # t.write('\n')
    # t.read_until('Output UT time format   [JD,CAL,BOTH]')
    # t.write('JD\n')
    # t.read_until('Output time digits  [MIN,SEC,FRACSEC]')
    # t.write('SEC\n')
    # t.read_until('Output R.A. format       [ HMS, DEG ]')
    # t.write('DEG\n')
    # t.read_until('Output high precision RA/DEC [YES,NO]')
    # t.write('YES\n')
    # t.read_until('Output APPARENT [ Airless,Refracted ]')
    # t.write('Airless\n')
    # t.read_until('Set units for RANGE output [ KM, AU ]')
    # t.write('AU\n')
    # t.read_until('Suppress RANGE_RATE output [ YES,NO ]')
    # t.write('YES\n')
    # t.read_until('Minimum elevation [ -90 <= elv <= 90]')
    # t.write('\n')

    # t.write('n\n')
    # t.read_until('Select table quantities [ <#,#..>, ?]')
    # t.write('1\n')
    # t.read_until('Output reference frame [J2000, B1950]')
    # t.write('J2000\n')
    # t.read_until('Time-zone correction   [ UT=00:00,? ]')
    # t.write('\n')
    # t.read_until('Output UT time format   [JD,CAL,BOTH]')
    # t.write('JD\n')
    # t.read_until('Output time digits  [MIN,SEC,FRACSEC]')
    # t.write('SEC\n')
    # t.read_until('Output R.A. format       [ HMS, DEG ]')
    # t.write('DEG\n')
    # t.read_until('Output high precision RA/DEC [YES,NO]')
    # t.write('YES\n')
    # t.read_until('Output APPARENT [ Airless,Refracted ]')
    # t.write('Airless\n')
    # t.read_until('Set units for RANGE output [ KM, AU ]')
    # t.write('AU\n')
    # t.read_until('Suppress RANGE_RATE output [ YES,NO ]')
    # t.write('YES\n')
    # t.read_until('Minimum elevation [ -90 <= elv <= 90]')
    # t.write('-90\n')

    #txt = t.read_until('Working ...', 10)
    #eph = t.read_until('>>> Select...', 60)

    txt = t.read_until('Ephemeris / PORT_LOGIN')
    header = t.read_until('$$SOE')
    eph = t.read_until('$$EOE')

    print('Header:', header)
    print('Eph:', eph)

    lines = eph.split('\n')
    print('first line:', lines[0])
    print('last line:', lines[-1])
    lines = lines[1:-1]
    print('first line:', lines[0])
    print('last line:', lines[-1])
    
    date,ra,dec = [],[],[]
    for line in lines:
        words = line.split()
        date.append(datetime.datetime.strptime(' '.join(words[:2]), '%Y-%b-%d %H:%M'))
        words = words[2:]
        ra.append(hmsstring2ra(' '.join(words[:3])))
        words = words[3:]
        dec.append(dmsstring2dec(' '.join(words[:3])))
    
    return date, ra, dec
    
    
def get_ephemerides_for_jds(bodyname, jds, debug=False):
    print('JDs:', jds)
    t = _horizons_login()
    ephs = []
    jd = jds[0]
    margin = 1. / (24.*3600.)
    
    print('getting JD', jd)
    t.write( # Body name; if found, it asks "Continue?"
             '%s\n\n' % bodyname +
             # [E]phemeris
             'E\n' +
             # [O]bserve, [E]lements, [V]ectors?
             'v\n' +
             '@0\n' +
             'eclip\n' +
             'JD %.9f\n' % jd +
             'JD %.9f\n' % (jd+margin) +
             '1h\n' +
             'n\nJ2000\n' +
             '1\n' + # corrections = NONE
             '2\n' + # AU/days
             'YES\n' + # CSV format
             'YES\n' + # label cartesian output?
             '3\n')
    txt = t.read_until('Working ...', 10)
    eph = t.read_until('>>> Select...', 60)
    if debug:
        print('Got eph: "%s"' % eph)
    ephs.append(eph)

    for jd in jds[1:]:
        print('getting JD', jd)
        t.write('A\nE\nv\n' +
                '\neclip\n' +
                'JD %.9f\n' % jd +
                'JD %.9f\n' % (jd+margin) +
                '1h\n' +
                '\n'
                )
        eph2 = t.read_until('>>> Select...', 60)
        ephs.append(eph2)
        if debug:
            print('Got eph: "%s"' % eph2)
        
    t.write('q\n')
    t.close()

    if debug:
        print('Got text:')
        for i,txt in enumerate(ephs):
            print('  ', txt)

            fn = 'txt%i' % i
            write_file(txt, fn)
            print('wrote', fn)

    EE = [Eph(txt=txt, linedelim='\r\n') for txt in ephs]
    return EE

if __name__ == '__main__':
    
    import optparse
    parser = optparse.OptionParser()
    parser.add_option('--start', default='2000-1-1',
                      help='Start date, YYYY-M-D, default %default')
    parser.add_option('--end', default='2021-1-1',
                      help='End date, YYYY-M-D, default %default')
    parser.add_option('--interval', default='1d',
                      help='Interval, default %default')
    parser.add_option('--body', default='C/2012 S1',
                      help='Solar system body, default %default')
    parser.add_option('--fits', default='ison-ephem.fits',
                      help='FITS table output filename, default %default')

    parser.add_option('--verbose', '-v', action='store_true', default=False,
                      help='Verbose mode?')
    
    opt,args = parser.parse_args()

    date0 = opt.start.split('-')
    if len(date0) != 3:
        print('Expected YYYY-M-D, got', opt.start)
        sys.exit(-1)
    date0 = [int(x, 10) for x in date0]
    date0 = datetime.datetime(*date0)
    print('Start date:', date0)
    jd0 = datetojd(date0)
    print('Start JD:', jd0)

    date1 = opt.end.split('-')
    if len(date1) != 3:
        print('Expected YYYY-M-D, got', opt.end)
        sys.exit(-1)
    date1 = [int(x, 10) for x in date1]
    date1 = datetime.datetime(*date1)
    print('End date:', date1)
    jd1 = datetojd(date1)
    print('End JD:', jd1)

    date,ra,dec = get_radec_for_jds(opt.body, jd0, jd1, debug=opt.verbose,
                                    interval=opt.interval)
    for d,rr,dd in zip(date, ra, dec):
        print('  ', d, rr, dd)

    T = fits_table()
    T.jd  = np.array([datetojd(d) for d in date])
    T.ra  = np.array(ra)
    T.dec = np.array(dec)
    T.writeto(opt.fits)
    print('Wrote', opt.fits)

    sys.exit(0)
    
    # import numpy as np
    # jd = 2454153.93624056 + np.random.normal(size=10)*300.
    # ephs = get_ephemerides_for_jds('GALEX', jd)
    # for j,E in zip(jd,ephs):
    #   E = E.entries[0]
    #   print
    #   print j
    #   print E.jd, E.lt, E.xyz
    
    jd0 = datetojd(datetime.datetime(2013,  9,  1))
    #jd1 = datetojd(datetime.datetime(2014,  3, 31))
    jd1 = datetojd(datetime.datetime(2014,  1,  1))
    print('jds', jd0, jd1)
    date,ra,dec = get_radec_for_jds('C/2012 S1', jd0, jd1, debug=True,
                                    interval='1h')
    
    for d,r,dd in zip(date, ra, dec):
        print('  ', d, r, dd)

    T = fits_table()
    T.jd  = np.array([datetojd(d) for d in date])
    T.ra  = np.array(ra)
    T.dec = np.array(dec)
    T.writeto('ison-ephem.fits')



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
from .common import *
from .dr8 import *

class DR9(DR8):

    def __init__(self, **kwargs):
        '''
        Useful kwargs:
        
        basedir : (string) - local directory where data will be stored.
        '''
        DR8.__init__(self, **kwargs)
        self.dasurl = 'http://data.sdss3.org/sas/dr9/boss/'

    def getDRNumber(self):
        return 9
        
    def _get_runlist_filename(self):
        return self._get_data_file('runList-dr9.par')


if __name__ == '__main__':
    sdss = DR9()
    rcfb = (2873, 3, 211, 'r')
    r,c,f,b = rcfb
    bandnum = band_index(b)
    sdss.retrieve('psField', *rcfb)
    psfield = sdss.readPsField(r,c,f)
    dg = psfield.getDoubleGaussian(bandnum, normalize=True)

    psf = psfield.getPsfAtPoints(bandnum, 2048/2., 1489./2.)

    import matplotlib
    matplotlib.use('Agg')
    import pylab as plt
    import numpy as np
    
    H,W = psf.shape
    cx,cy = (W/2, H/2)
    DX,DY = np.meshgrid(np.arange(W)-cx, np.arange(H)-cy)
    (a1,s1, a2,s2) = dg
    R2 = (DX**2 + DY**2)
    G = (a1 / (2.*np.pi*s1**2) * np.exp(-R2/(2.*s1**2)) +
         a2 / (2.*np.pi*s2**2) * np.exp(-R2/(2.*s2**2)))
    print('G sum', G.sum())
    print('psf sum', psf.sum())
    psf /= psf.sum()
    
    plt.clf()
    plt.subplot(2,2,1)
    ima = dict(interpolation='nearest', origin='lower')
    plt.imshow(psf, **ima)
    plt.subplot(2,2,2)
    plt.imshow(G, **ima)

    plt.subplot(2,2,3)
    plt.plot(psf[H/2,:], 'rs-', mec='r', mfc='none')
    plt.plot(G[H/2,:], 'gx-')

    plt.subplot(2,2,4)
    plt.semilogy(np.maximum(1e-6, psf[H/2,:]), 's-', mec='r', mfc='none')
    plt.semilogy(np.maximum(1e-6, G[H/2,:]), 'gx-')

    plt.savefig('psf1.png')


    
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from distutils.core import setup, Extension
import os
import sys

sys.path.append(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'util'))
from an_build_ext import an_build_ext

setup(cmdclass={'build_ext': an_build_ext},
      name='C utils for Astrometry.net SDSS routines',
      version = '1.0',
      author = 'Lang & Hogg',
      author_email = 'dstndstn@gmail.com',
      url = 'http://astrometry.net',
      ext_modules = [Extension('_cutils', ['cutils.i'], libraries=['m'])],
    )



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
from math import sqrt

if __name__ == '__main__':
    import matplotlib
    matplotlib.use('Agg')

import numpy as np
import pylab as plt

import astrometry.sdss
from astrometry.sdss import DR7
from astrometry.util.fits import fits_table

if __name__ == '__main__':
    sdss = DR7()

    testdata = os.path.join(os.path.dirname(astrometry.sdss.__file__),
                            'testdata')
    print('Using test data dir:', testdata)

    sdss.setBasedir(testdata)

    tsfield = sdss.readTsField(2830, 6, 398, 41)

    rband,iband = 2,3

    asr = tsfield.getAsTrans(rband)
    asi = tsfield.getAsTrans(iband)

    '''
$ listhead fpC-002830-r6-0398.fit

CRPIX1  = 1.02450000000000E+03 / Column Pixel Coordinate of Ref. Pixel
CRPIX2  = 7.44500000000000E+02 / Row Pixel Coordinate of Ref. Pixel
CRVAL1  = 1.79464261370000E+02 / RA at Reference Pixel
CRVAL2  = 5.33206072700000E+01 / DEC at Reference Pixel
CD1_1   = -8.4291038759718E-06 / RA  degrees per column pixel
CD1_2   = 1.09673333339369E-04 / RA  degrees per row pixel
CD2_1   = 1.09675673828125E-04 / DEC degrees per column pixel
CD2_2   = 8.42453629032368E-06 / DEC degrees per row pixel
'''

    print('CD at 0,0:', asr.cd_at_pixel(1024.5, 744.5))

    for x,y,color in [ (0, 0, 0),

                       (np.array([1,2,3]),
                        np.array([0,100,200]),
                        0),

                       (np.array([1,2,3]),
                        np.array([0,100,200]),
                        np.array([0,1,2])),
                       ]:
        print()
        print('Pixel x,y', x,y)
        print('color', color)
        rr,dr = asr.pixel_to_radec(x, y, color)
        ri,di = asi.pixel_to_radec(x, y, color)
        print('r-band RA,Dec:', rr,dr)
        print('i-band RA,Dec:', ri,di)
        rx,ry = asr.radec_to_pixel(rr, dr)
        ix,iy = asi.radec_to_pixel(ri, di)
        print('r-band x,y:', rx, ry)
        print('i-band x,y:', ix, iy)


    tsobj = fits_table(os.path.join(testdata, 'cut-tsObj-002830-6-0-0398.fit'))
    ra,dec = tsobj.ra, tsobj.dec
    #X,Y = tsobj.colc[:,rband], tsobj.rowc[:,rband]
    X,Y = tsobj.objc_colc, tsobj.objc_rowc
    print('ra,dec', ra.shape, dec.shape)
    print('x', X.shape)
    #rmag = tsobj.psfcounts[:,rband]
    #imag = tsobj.psfcounts[:,iband]
    rmag = tsobj.counts_model[:,rband]
    imag = tsobj.counts_model[:,iband]
    # According to http://www.sdss.org/dr7/dm/flatFiles/tsField.html,
    #   r,i,z fields, use this color:
    color = (rmag - imag)
    print('color:', color.min(), color.max())

    xx,yy = asr.radec_to_pixel(ra, dec, color)
    print('xx', xx.shape)
    rr,dd = asr.pixel_to_radec(X, Y, color)

    print('dxy', xx-X, yy-Y)
    #print 'dradec', ra-rr, dec-dd

    print('x RMS:', sqrt(np.mean((xx - X)**2)))
    print('y RMS:', sqrt(np.mean((yy - Y)**2)))
    print('RA RMS:', sqrt(np.mean((ra - rr)**2)))
    print('Dec RMS:', sqrt(np.mean((dec - dd)**2)))

    #I = np.argmax((xx - X)**2 + (yy - Y)**2)
    #print 'Biggest x,y deviant: row', I
    #tsobj[I].about()
    
    plt.clf()
    plt.plot(xx,yy, 'r.')
    plt.plot(X, Y,  'bo', mfc='none')
    plt.savefig('dxy.png')
    
    lo,hi = 1e-10, 10
    rng = np.log10(lo), np.log10(hi)
    plt.clf()
    plt.hist(np.log10(np.clip(np.abs(xx - X), lo, hi)), range=rng)
    plt.xlabel('log_10 ( x error )')
    plt.savefig('dx.png')

    plt.clf()
    plt.hist(np.log10(np.clip(np.abs(yy - Y), lo, hi)), range=rng)
    plt.xlabel('log_10 ( y error )')
    plt.savefig('dy.png')

    plt.clf()
    plt.hist(np.log10(np.clip(np.sqrt((yy - Y)**2 + (xx - X)**2), lo, hi)),
             range=rng)
    plt.xlabel('log_10 ( x,y error )')
    plt.savefig('dr.png')




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from astrometry.util.starutil_numpy import *
from astrometry.util.fits import *

'''
Updated Zwicky Catalog

https://www.cfa.harvard.edu/~falco/UZC/
https://www.cfa.harvard.edu/~falco/UZC/uzcJ2000.tab.gz
'''

rr,dd,nn = [],[],[]

for i,line in enumerate(open('uzcJ2000.tab')):
    if i < 20:
        continue
    #print 'line', line
    words = line.split('\t')
    #print 'words', words
    assert(len(words) == 14)

    ra = words[0]
    ra = ra[:2] + ' ' + ra[2:4] + ' ' + ra[4:]

    dec = words[1]
    dec = dec[:3] + ' ' + dec[3:5] + ' ' + dec[5:]

    #print 'ra,dec', ra,dec
    ra = hmsstring2ra(ra)
    dec = dmsstring2dec(dec)

    zname = words[8]
    
    #print 'ra,dec', ra,dec
    rr.append(ra)
    dd.append(dec)
    nn.append(zname)
    
T = tabledata()
T.ra = rr
T.dec = dd
T.zname = nn
T.writeto('uzc2000.fits')



from __future__ import print_function
from distutils.core import setup, Extension
from setuptools.command.install import install
import subprocess
import os
import sys
from glob import glob

class MyInstall(install):
    def run(self):
        print('MyInstall.run: calling "make -k"')
        subprocess.call(['make', '-k'])
        print('MyInstall.run: calling "make -k py"')
        subprocess.call(['make', '-k', 'py'])

        for cmd in ['make -k pyinstall',
                    'make -k install']:
            dirnm = self.install_base
            if dirnm is not None:
                cmd += ' INSTALL_DIR="%s"' % dirnm
            pybase = self.install_platlib
            if pybase is not None:
                pybase = os.path.join(pybase, 'astrometry')
                cmd += ' PY_BASE_INSTALL_DIR="%s"' % pybase
            py = sys.executable
            if py is not None:
                cmd += ' PYTHON="%s"' % py
            print('Running:', cmd)
            subprocess.call(cmd, shell=True)
            install.run(self)

class MyBuildExt(install):
    def run(self):
        print('MyBuildExt.run: calling "make -k"')
        subprocess.call(['make', '-k'])
        print('MyBuildExt.run: calling "make -k py"')
        subprocess.call(['make', '-k', 'py'])

setup(name='astrometry',
      version='git',
      author='Astrometry.net team',
      author_email='dstndstn@gmail.com',
      url='http://astrometry.net',
      cmdclass={'install': MyInstall, 'build_ext': MyBuildExt},
      packages=['astrometry'],
      package_dir={'astrometry':''},
      )



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
from astrometry.util.file import *

'''

Stages: a utility for saving and resuming computation, savings
intermediate results as pickle files.


'''

class CallGlobal(object):
    def __init__(self, pattern, globals, *args, **kwargs):
        self.pat = pattern
        self.args = args
        self.kwargs = kwargs
        self.globals = globals
    def getfunc(self, stage):
        func = self.pat % stage
        func = eval(func, self.globals)
        return func
    def getkwargs(self, stage, **kwargs):
        kwa = self.kwargs.copy()
        kwa.update(kwargs)
        return kwa
    def __call__(self, stage, **kwargs):
        func = self.getfunc(stage)
        kwa = self.getkwargs(stage, **kwargs)
        return func(*self.args, **kwa)

class CallGlobalTime(CallGlobal):
    def __call__(self, stage, **kwargs):
        from astrometry.util.ttime import Time
        from datetime import datetime
        t0 = Time()
        print('Running stage', stage, 'at', datetime.now().isoformat())
        rtn = super(CallGlobalTime, self).__call__(stage, **kwargs)
        t1 = Time()
        print('Stage', stage, ':', t1-t0)
        print('Stage', stage, 'finished:', datetime.now().isoformat())
        return rtn

def runstage(stage, picklepat, stagefunc, force=[], forceall=False, prereqs={},
             update=True, write=True, initial_args={}, **kwargs):
    '''
    Run to a given *stage*.

    Each stage is a function.
    
    Each stage takes a dict and returns None or a dict.

    Each stage (except the first one!) has a prerequisite stage.

    Results after running each stage can be written to a pickle file,
    and later read instead of running the stage.

    Parameters
    ----------
    stage - int or string
        Stage name
    picklepat - string
        Filename pattern to which pickle data should be written
    stagefunc - function
        Function to call for each stage.
    force - list of stage name/number
        List of stages to run, ignoring existing pickle files
    forceall - boolean
        Force running all stages; ignore all existing pickle files
    prereqs - dict of stage->stage mappings
        Defines the dependencies between stages
    update - boolean
        Update prerequiste dict with results before writing out pickle
    write - boolean, or list of stages
        Write a pickle for this stage / all stages?
    initial_args - dict
        Arguments to pass to the first stage
    kwargs - dict
        Keyword args added to the prerequisites before running each stage
    '''
    # NOTE, if you add or change args here, be sure to update the recursive
    # "runstage" call below!!
    print('Runstage', stage)

    try:
        pfn = picklepat % stage
    except:
        pfn = picklepat % dict(stage=stage)
    
    if os.path.exists(pfn):
        if forceall or stage in force:
            print('Ignoring pickle', pfn, 'and forcing stage', stage)
        else:
            print('Reading pickle', pfn)
            try:
                R = unpickle_from_file(pfn)
                return R
            except Exception as e:
                print('Failed to read pickle file', pfn, ':', e, '; re-running stage')

    try:
        prereq = prereqs[stage]
    except KeyError:
        prereq = stage - 1

    if prereq is None:
        P = initial_args
    else:
        P = runstage(prereq, picklepat, stagefunc,
                     force=force, forceall=forceall, prereqs=prereqs, update=update,
                     write=write, initial_args=initial_args, **kwargs)

    #P.update(kwargs)
    Px = P.copy()
    Px.update(kwargs)

    print('Running stage', stage)
    # print('Prereq keys:', P.keys())
    # print('Adding kwargs keys:', kwargs.keys())
    # print('Combined keys:', Px.keys())

    R = stagefunc(stage, **Px)
    print('Stage', stage, 'finished')
    # if R is not None:
    #     print('Result keys:', R.keys())

    if update:
        if R is not None:
            P.update(R)
        R = P

    if not write:
        pass
    elif (write is True or stage in write):
        print('Saving pickle', pfn)
        #print('Pickling keys:', R.keys())
        # Create directory, if necessary.
        dirnm = os.path.dirname(pfn)
        if len(dirnm) and not os.path.exists(dirnm):
            try:
                os.makedirs(dirnm)
            except:
                pass
        tempfn = os.path.join(dirnm, 'tmp-' + os.path.basename(pfn))
        pickle_to_file(R, tempfn)
        os.rename(tempfn, pfn)
        print('Saved', pfn)
    return R



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
import os
import sys
import logging
from optparse import OptionParser
import numpy as np
from astrometry.util.fits import fits_table

# Returns a numpy array of booleans
def hist_remove_lines(x, binwidth, binoffset, logcut):
    bins = -binoffset + np.arange(0, max(x)+binwidth, binwidth)
    (counts, thebins) = np.histogram(x, bins)

    # We're ignoring empty bins.
    occupied = np.nonzero(counts > 0)[0]
    noccupied = len(occupied)
    if noccupied == 0:
        return np.array([True] * len(x))
    k = (counts[occupied] - 1) 
    mean = sum(k) / float(noccupied)
    logpoisson = k*np.log(mean) - mean - np.array([sum(np.arange(kk)) for kk in k])
    badbins = occupied[logpoisson < logcut]
    if len(badbins) == 0:
        return np.array([True] * len(x))

    badleft = bins[badbins]
    badright = badleft + binwidth

    badpoints = sum(np.array([(x >= L)*(x < R) for (L,R) in
                              zip(badleft, badright)]), 0)
    return (badpoints == 0)

def removelines(infile, outfile, xcol='X', ycol='Y', ext=1, cut=None, **kwargs):
    if cut is None:
        cut = 100
    T = fits_table(infile, lower=False)
    if len(T) == 0:
        print('removelines.py: Input file contains no sources.')
        T.writeto(outfile)
        return 0
    
    ix = hist_remove_lines(T.get(xcol), 1, 0.5, logcut=-cut)
    iy = hist_remove_lines(T.get(ycol), 1, 0.5, logcut=-cut)
    Norig = len(T)
    T.cut(ix * iy)
    print('removelines.py: Removed %i sources' % (Norig - len(T)))
    #header.add_history('This xylist was filtered by the "removelines.py" program')
    #header.add_history('to remove horizontal and vertical lines of sources')
    #header['REMLINEN'] = (len(x) - len(xc), 'Number of sources removed by "removelines.py"')
    T.writeto(outfile)
    return 0

def main():
    parser = OptionParser(usage='%prog [options] <input-xylist> <output-xylist>')

    parser.add_option('-X', dest='xcol', help='Name of X column in input table')
    parser.add_option('-Y', dest='ycol', help='Name of Y column in input table')
    parser.add_option('-s', dest='cut', type='float', help='Significance level to cut at (default 100)')
    parser.add_option('-e', dest='ext', type=int, help='FITS extension to read',
                      default=1)
    parser.set_defaults(xcol='X', ycol='Y', cut=None)

    (opt, args) = parser.parse_args()
    
    if len(args) != 2:
        parser.print_help()
        print()
        print('Got arguments:', args)
        sys.exit(-1)

    infile = args[0]
    outfile = args[1]
    return removelines(infile, outfile, xcol=opt.xcol, ycol=opt.ycol,
                       cut=opt.cut, ext=opt.ext)

if __name__ == '__main__':
    sys.exit(main())



import os
from distutils.core import setup, Extension

from numpy.distutils.misc_util import get_numpy_include_dirs
numpy_inc = get_numpy_include_dirs()

inc = ['include', 'include/astrometry', 'util', 'qfits-an']

libkd_srcs = [
    'pyspherematch.c',
    'dualtree.c', 'dualtree_rangesearch.c', 'dualtree_nearestneighbour.c',
    'kdtree.c', 'kdtree_dim.c', 'kdtree_mem.c',
    'kdtree_fits_io.c',
    'kdint_ddd.c',
    'kdint_fff.c',
    'kdint_ddu.c',
    'kdint_duu.c',
    'kdint_dds.c',
    'kdint_dss.c',
    ]
util_srcs = [
    'ioutils.c', 'bl.c', 'mathutil.c', 'fitsioutils.c', 'fitsbin.c',
    'an-endian.c', 'fitsfile.c', 'log.c', 'errors.c', 'tic.c',
    ]

qfits_srcs = [
    'anqfits.c', 'qfits_tools.c', 'qfits_table.c', 'qfits_float.c',
    'qfits_error.c', 'qfits_time.c', 'qfits_card.c', 'qfits_header.c',
    'qfits_rw.c', 'qfits_memory.c', 'qfits_convert.c', 'qfits_byteswap.c',
    ]

srcs = ([os.path.join('libkd',x) for x in libkd_srcs] +
        [os.path.join('util', x) for x in util_srcs]  +
        [os.path.join('qfits-an', x) for x in qfits_srcs]
    )

    
ext = Extension('astrometry.libkd.spherematch_c',
                sources = srcs,
                include_dirs = numpy_inc + inc,
                )

setup(name='libkd',
      version='1.0',
      author='Dustin Lang, Keir Mierle',
      author_email='dstndstn@gmail.com',
      url='http://astrometry.net',
      ext_modules=[ext],
      py_modules=['astrometry.libkd.spherematch',
                  'astrometry.util.starutil_numpy',
                  'astrometry.__init__',
                  'astrometry.util.__init__',
                  'astrometry.libkd.__init__',
                  ],
      package_dir={'astrometry':''},
      data_files=[('tests', ['libkd/test_spherematch.py']),],
    )




from __future__ import absolute_import
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from .common import *
from .dr9 import *

class DR10(DR9):

    def __init__(self, **kwargs):
        '''
        Useful kwargs:
        
        basedir : (string) - local directory where data will be stored.
        '''
        super(DR9, self).__init__(**kwargs)
        self.dasurl = 'http://data.sdss3.org/sas/dr10/boss/'

    def getDRNumber(self):
        return 10
        
    def _get_runlist_filename(self):
        return self._get_data_file('runList-dr10.par')




#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import sys
from optparse import OptionParser

try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
from numpy import *

from astrometry.util.fits import *
from astrometry.util.healpix import *
from astrometry.util.starutil_numpy import *
from astrometry.util.usnob_cuts import *

def get_usnob_sources(ra, dec, radius=1, basefn=None):
    usnob_nside = 9

    if basefn is None:
        usnob_pat = 'usnob10_hp%03i.fits'
    else:
        usnob_pat = basefn

    usnobhps = healpix_rangesearch(ra, dec, radius, usnob_nside)
    print('USNO-B healpixes in range:', usnobhps)
    allU = None
    for hp in usnobhps:
        usnobfn = usnob_pat % hp
        print('USNOB filename:', usnobfn)
        U = table_fields(usnobfn)
        I = (degrees_between(ra, dec, U.ra, U.dec) < radius)
        print('%i USNOB stars within range.' % sum(I))
        U = U[I]
        if allU is None:
            allU = U
        else:
            allU.append(U)
    return allU

if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <ra> <dec> <output-filename>')

    parser.add_option('-r', dest='radius', type='float', help='Search radius, in deg (default 1 deg)')
    parser.add_option('-b', dest='basefn', help='Base filename of USNO-B FITS files (default: usnob10_hp%03i.fits)')
    parser.set_defaults(radius=1.0, basefn=None)

    (opt, args) = parser.parse_args()
    if len(args) != 3:
        parser.print_help()
        print()
        print('Got extra arguments:', args)
        sys.exit(-1)

    # parse RA,Dec.
    ra = float(args[0])
    dec = float(args[1])
    outfn = args[2]

    # ugh!
    opts = {}
    for k in ['radius', 'basefn']:
        opts[k] = getattr(opt, k)

    X = get_usnob_sources(ra, dec, **opts)
    print('Got %i USNO-B sources.' % len(X))

    print('Applying cuts...')
    I = usnob_apply_cuts(X)
    X = X[I]
    print(len(X), 'pass cuts')

    usnob_compute_average_mags(X)

    print('Writing to', outfn)
    X.write_to(outfn)
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import simplejson
import re

lines = open('brightstars-data.c').readlines()
l = ''.join(lines[3:-2]).replace('{','[').replace('}',']')

def replace_unicode(match):
    c1 = match.group(1)
    c2 = match.group(2)
    #s = eval('\\x%s\\x%s'
    s = '"\\x%s\\x%s"' % (c1, c2)
    #print 's', s
    s = eval(s)
    #print 's', s
    d = s.decode('utf8')
    return d + ' '

l = re.sub(r'\\x(..)\\x(..)""', replace_unicode, l)
#l = re.sub(r'\\x(..)\\x(..)""', r'\u\1\2 ', l)
#l = re.sub(r'\\x(..)\\x(..)""', r'\x\1\x\2 ', l)
#l = l.decode('utf8')

l = '[' + l + '0 ]'
print(l)

j = simplejson.loads(l)
j = j[:-1]
print(j)

nm, nm2, rr, dd = [],[],[],[]
vmag = []
for n1,n2,r,d,mag in j:
    #print 'n1', n1
    j = simplejson.dumps(n1)
    #print 'json:', j
    #j = str(j)
    #print '  ->', j
    nm.append(j.replace('"', ''))
    nm2.append(str(n2))
    vmag.append(float(mag))
    rr.append(r)
    dd.append(d)

from astrometry.util.fits import *
import numpy as np

T = tabledata()
T.name1 = np.array(nm)
T.name2 = np.array(nm2)
T.vmag = np.array(vmag)
T.ra = np.array(rr)
T.dec = np.array(dd)
T.about()
T.writeto('brightstars.fits')



# celestial_mechanics.py
#   celestial mechanics utilities for exoplanet ephemerides
#
# intellectual property:
#   Copyright 2009 David W. Hogg.  All rights reserved.
#   Licensed under a 3-clause BSD style license - see LICENSE
#
# comments:
#   - Written for clarity, not speed.  The code is intended to be human-
#     readable.
#
# bugs:
#   - Need to make Fourier expansion functions.
#    - What to do if e is close to 1.0 in eccentric_anomaly
#
from __future__ import print_function

from math import pi
import unittest
import sys

import numpy
import numpy as np
from numpy import *
from scipy.special import jn
import matplotlib.pyplot as plt

from astrometry.util.starutil_numpy import *

ihat = array([1.,0.,0.])
jhat = array([0.,1.,0.])
khat = array([0.,0.,1.])
default_tolerance = 1e-15 # (radians) don't set to zero or less
default_maximum_iteration = 1000 # should never hit this limit!
default_order = 32
default_K = 1.0

(Equinox, Solstice, EclipticPole) = ecliptic_basis()

c_au_per_yr = 63239.6717 # google says

#GM_sun = 2.9591310798672560E-04 #AU^3/d^2
# Google says mass of the sun * G = 39.4775743 (au^3) / (yr^2)
GM_sun = 39.4775743 # AU^3 / yr^2

def norm1d(x):
    assert(len(x.shape) == 1)
    return np.sqrt(np.sum(x**2))

def deg2rad(x):
    return x * pi/180.
    #return radians(x)

def orbital_elements_to_ss_xyz(E, observer=None, light_travel=True):
    (a,e,i,Omega,pomega,M,GM) = E
    # ugh, it's hard to be units-agnostic.
    # we just assert here so we have to think about this!
    # This means:
    #  distances in AU
    #  angles in radians
    #  times in years
    assert(GM == GM_sun)
    if light_travel:
        assert(observer is not None)
        # orbital angular velocity  [radians/yr]
        meanfrequency = np.sqrt(GM / a**3)
    # Correct for light-time delay.
    # dM: [radians]
    dM = 0.
    lastdM = dM
    dx = None
    for ii in range(100):
        (x,v) = phase_space_coordinates_from_orbital_elements(
            a,e,i,Omega,pomega,M-dM,GM)
        if not light_travel:
            if observer is not None:
                dx = x - observer
            break
        dx = x - observer
        # light-travel distance [AU]
        r = norm1d(dx)
        # light-travel time [yr]
        travel = r / c_au_per_yr
        #print 'light-travel time:', travel, 'yr, or', travel*365.25, 'd'
        # light travel in angular periods [radians]
        dM = travel * meanfrequency
        if abs(lastdM - dM) < 1e-12:
            break
        lastdM = dM
    if ii == 99:
        print('Warning: orbital_elements_to_ss_xyz: niters', ii)
    return x,dx

def orbital_elements_to_xyz(E, observer, light_travel=True, normalize=True):
    (x,dx) = orbital_elements_to_ss_xyz(E, observer, light_travel)
    if normalize:
        dx /= norm1d(dx)
    edx = dx[0] * Equinox + dx[1] * Solstice + dx[2] * EclipticPole
    return edx

# E = (a,e,i,Omega,pomega,M, GM)
# observer = 3-vector
# light_travel: correct for light travel time?
# Returns RA,Dec in degrees.
def orbital_elements_to_radec(E, observer, light_travel=True):
    xyz = orbital_elements_to_xyz(E, observer, light_travel)
    return xyztoradec(xyz)

# convert orbital elements into vectors in the plane of the orbit.
def orbital_vectors_from_orbital_elements(i, Omega, pomega):
    ascendingnodevector = np.cos(Omega) * ihat + np.sin(Omega) * jhat
    tmpydir= np.cross(khat, ascendingnodevector)
    zhat= np.cos(i) * khat - np.sin(i) * tmpydir
    tmpydir= np.cross(zhat, ascendingnodevector)
    xhat= np.cos(pomega) * ascendingnodevector + np.sin(pomega) * tmpydir
    yhat = np.cross(zhat, xhat)
    return (xhat, yhat, zhat)

def position_from_orbital_vectors(xhat, yhat, a, e, M):
    E = eccentric_anomaly_from_mean_anomaly(M, e)
    cosE = np.cos(E)
    sinE = np.sin(E)
    b = a*np.sqrt(1. - e**2)
    x =  a * (cosE - e)  * xhat + b * sinE        * yhat
    return x

# convert orbital elements to phase-space coordinates
#  a       - semi-major axis (length units)
#  e       - eccentricity
#  i       - inclination (rad)
#  Omega   - longitude of ascending node (rad)
#  pomega  - argument of periapsis (rad)
#  M       - mean anomaly (rad)
#  GM      - Newton's constant times central mass (length units cubed over time units squared)
#  return  - (x,v)
#            position, velocity (length units, length units per time unit)
def phase_space_coordinates_from_orbital_elements(a, e, i, Omega, pomega, M, GM):
    (xhat, yhat, zhat) = orbital_vectors_from_orbital_elements(i, Omega, pomega)
    # [radians/yr]
    dMdt = np.sqrt(GM / a**3)
    E = eccentric_anomaly_from_mean_anomaly(M, e)
    cosE = np.cos(E)
    sinE = np.sin(E)
    # [radians/yr]
    dEdt = 1.0 / (1.0 - e * cosE) * dMdt
    # [AU]
    b = a*np.sqrt(1. - e**2)
    x =  a * (cosE - e)  * xhat + b * sinE        * yhat
    # [AU/yr]
    v = -a * sinE * dEdt * xhat + b * cosE * dEdt * yhat
    return (x, v)

class UnboundOrbitError(ValueError):
    pass

def potential_energy_from_position(x, GM):
    return -1. * GM / norm1d(x)

def energy_from_phase_space_coordinates(x, v, GM):
    return 0.5 * np.dot(v, v) + potential_energy_from_position(x, GM)

# convert phase-space coordinates to orbital elements
#  x       - position (3-vector, length units)
#  v       - velocity (3-vector, length units per time unit)
#  GM      - Newton's constant times central mass (length units cubed over time units squared)
#  return  - (a, e, i, Omega, pomega, M)
#          - see "phase_space_coordinates" for definitions
def orbital_elements_from_phase_space_coordinates(x, v, GM):
    energy = energy_from_phase_space_coordinates(x, v, GM)
    if energy > 0:
        raise UnboundOrbitError('orbital_elements_from_phase_space_coordinates: Unbound orbit')

    angmom = np.cross(x, v)
    zhat = angmom / norm1d(angmom)
    evec = np.cross(v, angmom) / GM - x / norm1d(x)
    e = norm1d(evec)
    if e == 0:
        # by convention:
        xhat = np.cross(jhat, zhat)
        xhat /= norm1d(xhat)
    else:
        xhat = evec / e
    yhat = np.cross(zhat, xhat)
    a = -0.5 * GM / energy
    i = np.arccos(angmom[2] / norm1d(angmom))
    if i == 0:
        Omega = 0.0
    else:
        Omega = np.arctan2(angmom[1], angmom[0]) + 0.5 * pi
        if Omega < 0:
            Omega += 2.*pi
        if i < 0:
            i *= -1.
            Omega += pi
    cosOmega = cos(Omega)
    sinOmega = sin(Omega)
    if e == 0:
        pomega = 0. - Omega
    else:
        pomega = np.arccos(min(1.0, (evec[0] * cosOmega + evec[1] * sinOmega) / e))
    horriblescalar = ( sinOmega * evec[2] * angmom[0]
             - cosOmega * evec[2] * angmom[1]
             + cosOmega * evec[1] * angmom[2]
             - sinOmega * evec[0] * angmom[2])
    if horriblescalar < 0.:
        pomega = 2.0 * pi - pomega
    if pomega < 0.0:
        pomega += 2.0 * pi
    if pomega > 2.0 * pi:
        pomega -= 2.0 * pi
    f = np.arctan2(np.dot(yhat, x), np.dot(xhat, x))
    M = mean_anomaly_from_true_anomaly(f, e)
    if M < 0:
        M += 2.*pi
    return (a, e, i, Omega, pomega, M)

# convert eccentric anomaly to mean anomaly
#  E       - eccentric anomaly (radians)
#  e       - eccentricity
#  return  - mean anomaly (radians)
def mean_anomaly_from_eccentric_anomaly(E, e):
    return (E - e * np.sin(E))

def mean_anomaly_from_true_anomaly(f, e):
    return mean_anomaly_from_eccentric_anomaly(eccentric_anomaly_from_true_anomaly(f, e), e)

# convert mean anomaly to eccentric anomaly
#  M       - [array of] mean anomaly (radians)
#  e       - eccentricity
#  [tolerance - read the source]
#  [maximum_iteration - read the source]
#  return  - eccentric anomaly (radians)
def eccentric_anomaly_from_mean_anomaly(M, e, tolerance = default_tolerance,
              maximum_iteration = default_maximum_iteration, verbose=False):
    E = M + e * np.sin(M)
    iteration = 0
    deltaM = 100.0
    while (iteration < maximum_iteration) and (abs(deltaM) > tolerance):
        deltaM = (M - mean_anomaly_from_eccentric_anomaly(E, e))
        E = E + deltaM / (1. - e * cos(E))
        iteration += 1
    if verbose: print('eccentric anomaly iterations:',iteration)
    return E

def eccentric_anomaly_from_true_anomaly(f, e):
    E = np.arccos((np.cos(f) + e) / (1.0 + e * np.cos(f)))
    E *= (np.sign(np.sin(f)) * np.sign(np.sin(E)))
    return E

# convert eccentric anomaly to true anomaly
#  E       - eccentric anomaly (radians)
#  e       - eccentricity
#  return  - true anomaly (radians)
def true_anomaly_from_eccentric_anomaly(E, e):
    f = np.arccos((np.cos(E) - e) / (1.0 - e * np.cos(E)))
    f *= (np.sign(np.sin(f)) * np.sign(np.sin(E)))
    return f

# compute radial velocity
#  K       - radial velocity amplitude
#  f       - true anomaly (radians)
#  e       - eccentricity
#  pomega  - eccentric longitude (radians)
#  return  - radial velocity (same units as K)
def radial_velocity(K, f, e, pomega):
    return K * (np.sin(f + pomega) + e * np.sin(pomega))

# compute radial velocity
#  K       - radial velocity amplitude
#  M       - mean anomaly (radians)
#  e       - eccentricity
#  pomega  - eccentric longitude (radians)
#  return  - radial velocity (same units as K)
def radial_velocity_from_M(K, M, e, pomega):
    E = M + e*np.sin(M)
    term1 = np.cos(pomega) * np.sqrt(1 - e**2) * np.sin(E) / (1 - e*np.cos(E))
    term2 = np.sin(pomega) * (np.cos(E) - e) / (1 - e*np.cos(E))
    term3 = e*np.sin(pomega)
    return K * (term1 + term2 + term3)

# compute radial velocity using a truncated Fourier series
#  K       - radial velocity amplitude
#  M       - mean anomaly (radians) APW: you may want to change this input
#  e       - eccentricity
#  pomega  - eccentric longitude (radians)
#  phi       - phase
#  [order  - read the source]
#  return  - radial velocity (same units as K)
def radial_velocity_fourier_series(K, M, e, pomega, phi, order=default_order):
    vr = 0.0
    for n in arange(0, order+1, 1):
        vr += K*(fourier_coeff_A(n, pomega, phi, e) * np.cos(n*(M-phi)) \
            + fourier_coeff_B(n, pomega, phi, e) * np.sin(n*(M-phi)))
    return vr

# the following is based on the naming convention in Itay's notes on Fourier analysis
#  - fourier_coeff_A and fourier_coeff_B are the actual coefficients in the series
#  - aprime and bprime are just used to simplify the code, and break it up to make it more readable

#  n       - order of the coefficient
#  e       - eccentricity
def aprime(n,e):
    return np.sqrt(1. - e**2)*( ((np.sqrt(1. - e**2) - 1)/e)*jn(n,n*e) + jn(n-1,n*e))

def bprime(n,e):
    return np.sqrt(1. - e**2)*( ((np.sqrt(1. - e**2) - 1)/e)*jn(n,n*e) - jn(n-1,n*e))

def fourier_coeff_A(n, pomega, phi, e):
    return 0.5 * (aprime(n,e) * np.sin(pomega + n * phi) + bprime(n,e) * np.sin(pomega - n * phi))

def fourier_coeff_B(n, pomega, phi, e):
    return 0.5 * (aprime(n,e) * np.cos(pomega + n * phi) - bprime(n,e) * np.cos(pomega - n * phi))

# APW: adjust function call as necessary
#  return  - amplitudes as a list of tuples (An,Bn)
def radial_velocity_fourier_amplitudes(K, phi, e, pomega, order=default_order):
    amplitudes = []
    for n in range(order):
        amplitudes.append((fourier_coeff_A(n, pomega, phi, e), fourier_coeff_B(n, pomega, phi, e)))
    return amplitudes

def eccentricity_from_fourier_amplitudes(amplitudes):
    K = np.sqrt(amplitude[0]**2 + amplitude[1]**2)
    phi = np.arctan(amplitude[1] / amplitude[0]) # WRONG?
    e = 0 # WRONG
    pomega = 0 # WRONG
    return (K, phi, e, pomega)

    

# some functional testing
if __name__ == '__main__':
    from test_celestial_mechanics import *
    #suite = unittest.TestLoader().loadTestsFromTestCase(TestOrbitalElements)

    suite = unittest.TestSuite()
    #suite.addTest(TestOrbitalElements('testEdgeCases'))
    suite.addTest(TestOrbitalElements('testAgainstJPL_2'))

    unittest.TextTestRunner(verbosity=2).run(suite)
    import sys
    sys.exit(0)
    
    # -- test Earth ephemeris against JPL at Holmes' closest approach
    # -- test Holmes against JPL         ---------''------------
    # -- test direction2radec()

    try:
        arg1 = sys.argv[1]
    except IndexError:
        arg1 = 0
    
    for e in [0.01, 0.1, 0.9, 0.99, 0.999]:
        print('eccentricity:', e)
        M = arange(-3.16,-3.14,0.001) # easy
        #M = arange(-10., 10., 2.1)    # easy
        #M = arange(-0.01,0.01,0.001)  # hard
        print('mean anomaly input:', M)
        E = eccentric_anomaly_from_mean_anomaly(M, e, verbose=True)
        print('eccentric anomaly output:', E)
        f = true_anomaly_from_eccentric_anomaly(E, e)
        print('true anomaly output:', f)
        M2 = mean_anomaly_from_eccentric_anomaly(E, e)
        print('round-trip error:', M2 - M)
    
    if arg1 == "plot":
        # This code will do the plotting:
        range_min = 0.0
        range_max = 20.0
        step_size = 0.01
        phase = 0.0
        pomegas = [0.0, pi/5., 3.*pi/5., 5.*pi/5., 7.*pi/5., 9.*pi/5.]
        pomegas_str = ["pomega = $0.0$", "pomega = $\pi/5$", "pomega = $3\pi/5$", "pomega = $\pi$", \
            "pomega = $7\pi/5$", "pomega = $9\pi/5$"]
        eccens = [0.01, 0.1, 0.5, 0.9, 0.99]
        orders = [2, 4, 8, 16, 32]
        M = arange(range_min,range_max,step_size)
        
        for n in orders:
            for e in eccens:
                i = 1
                for pomega in pomegas:
                    plt.clf()
                    plt.suptitle(pomegas_str[i-1] + ", e = %.2f, n = %i" % (e,n))
                    plt.subplot(211)
                    plt.plot(M, radial_velocity_fourier_series(default_K, M, e, pomega, phase, order=n), 'r')
                    plt.plot(M, radial_velocity_from_M(default_K, M, e, pomega), 'k--') 
                    plt.axis([range_min,range_max,-(default_K+1.),default_K+1.])
                    plt.subplot(212)
                    plt.plot(M, radial_velocity_from_M(default_K, M, e, pomega) \
                        - radial_velocity_fourier_series(default_K, M, e, pomega, phase, order=n), 'k')
                    plt.savefig("celestial_mechanics_plots/pomega_%i_e_%.2f_n_%i.png" % (i,e,n))
                    i += 1



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from xml.dom import minidom, Node
import numpy as np
from astrometry.util.fits import *

def siap_parse_result(fn=None):
    dom1 = minidom.parse(fn)

    tables = dom1.getElementsByTagName('TABLE')
    assert(len(tables) == 1)
    table = tables[0]
    
    fields = table.getElementsByTagName('FIELD')
    print('%i fields' % len(fields))
    fieldnames = []
    fieldtypes = []
    fieldparser = []
    fieldisarray = []
    for f in fields:
        name = f.getAttribute('name').lower().replace('[]', '')
        print('field:', name, end=' ')
        ftype = f.getAttribute('datatype').lower()
        print('(%s)' % ftype)
        farray = f.hasAttribute('arraysize')

        ftmap = {'int':int,
                 'double':float,
                 'float':float,
                 'char':str,
                 }

        fieldnames.append(name)
        fieldtypes.append(ftype)
        fieldparser.append(ftmap.get(ftype))
        fieldisarray.append(farray)

    data = table.getElementsByTagName('TABLEDATA')
    assert(len(data) == 1)
    data = data[0]

    rows = data.getElementsByTagName('TR')
    print('%i rows' % len(rows))

    datarows = []
    for r in rows:
        cols = r.getElementsByTagName('TD')
        assert(len(cols) == len(fields))
        datacol = []
        for c,ft,fp,fa in zip(cols, fieldtypes, fieldparser, fieldisarray):
            if len(c.childNodes) == 0:
                # blank
                if ft in ['double']:
                    datacol.append(np.nan)
                elif ft in ['int']:
                    datacol.append(0)
                else:
                    datacol.append('')
                continue
            assert(c.firstChild)
            c = c.firstChild
            # print('Node:', c)
            # print('Node value:', c.nodeValue)
            assert(c.nodeType in [Node.TEXT_NODE, Node.CDATA_SECTION_NODE])
            c = c.nodeValue
            datum = None
            if fa and ft in ['int','double']:
                #elements = c.split(',')
                c = c.replace(',', ' ')
                elements = c.split()
                datum = np.array([fp(x) for x in elements])
            elif fp:
                datum = fp(c)
            else:
                datum = c
            datacol.append(datum)
        datarows.append(datacol)

    t = tabledata()
    for i,f in enumerate(fieldnames):
        t.set(str(f), np.array([r[i] for r in datarows]))
    t._length = len(datarows)

    return t



from __future__ import print_function
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import os
import sys
# add .. to pythonpath
path = os.path.abspath(__file__)
sys.path.append(os.path.dirname(os.path.dirname(path)))

from distutils.core import setup, Extension
from numpy.distutils.misc_util import get_numpy_include_dirs

numpy_inc = get_numpy_include_dirs()

def strlist(s, split=' '):
    lst = s.split(split)
    lst = [i.strip() for i in lst]
    lst = [i for i in lst if len(i)]
    return lst

link = ' '.join([os.environ.get('LDFLAGS', ''),
                 os.environ.get('LDLIBS', ''),])
link = strlist(link)
objs = strlist(os.environ.get('SLIB', ''))
inc = strlist(os.environ.get('INC', ''), split='-I')
cflags = strlist(os.environ.get('CFLAGS', ''))

print(('link:', link))
print(('objs:', objs))
print(('inc:', inc))
print(('cflags:', cflags))

c_swig_module = Extension('_solver',
                          sources = ['solver.i'],
                          include_dirs = numpy_inc + inc + ['.'],
                          extra_objects = objs,
                          extra_compile_args = cflags,
                          extra_link_args=link,
                          depends=objs,
                          swig_opts=['-I'+d for d in inc] + cflags,
    )

setup(#cmdclass={'build_ext': an_build_ext},
      name = 'Access to Astrometry.net solver/ in python',
      version = '1.0',
      description = '',
      author = 'Astrometry.net (Dustin Lang)',
      author_email = 'dstndstn@gmail.com',
      url = 'http://astrometry.net',
      py_modules = [ 'solver' ],
      ext_modules = [c_swig_module])




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import sys
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
import numpy

from astrometry.util.fits import pyfits_writeto

def is_sdss_idr(hdu):
    hdr = hdu.header
    return ((hdr.get('SIMPLE', True) == False)
            and ('SDSS' in hdr)
            and ('UNSIGNED' in hdr))
            #and hdr.get('SDSS', None) == pyfits.UNDEFINED
            #and hdr.get('UNSIGNED', None) == pyfits.UNDEFINED)

def is_sdss_idr_file(infile):
    p = pyfits.open(infile)
    rtn = is_sdss_idr(p[0])
    p.close()
    return rtn

# Takes a pyfits HDU object (which is unchanged) and returns a new
# pyfits object containing the fixed file.
def fix_sdss_idr(hdu):
    hdr = hdu.header.copy()

    if hdr.get('SIMPLE', True):
        print('SIMPLE = T: not an SDSS idR file.')
        return hdu
    print('Setting SIMPLE = True')
    hdr.remove('SIMPLE')
    hdr.set('SIMPLE', True, 'FITS compliant (via fix-sdss-idr.py)')

    if 'SDSS' in hdr:
        print('Setting SDSS = True')
        hdr.remove('SDSS')
        hdr.set('SDSS', True, 'SDSS (via fix-sdss-idr.py)')
    else:
        print('No SDSS header card: not an SDSS idR file.')
        return hdu

    if 'UNSIGNED' in hdr:
        print('Setting UNSIGNED = True')
        hdr.remove('UNSIGNED')
        hdr.set('UNSIGNED', True, 'SDSS unsigned ints')
    else:
        print('No UNSIGNED header card: not an SDSS idR file.')
        return hdu

    #hdr._updateHDUtype()
    #newhdu = hdr._hdutype(data=pyfits.DELAYED, header=hdr)
    newhdu = pyfits.PrimaryHDU(data=pyfits.DELAYED, header=hdr)
    ## HACK - encapsulation violation
    newhdu._file = hdu._file
    #newhdu._ffile = hdu._ffile
    newhdu._datLoc = hdu._datLoc

    newhdu.data = newhdu.data.astype(numpy.int32)
    newhdu.data[newhdu.data < 0] += 2**16
    print('data type:', newhdu.data.dtype)
    print('data range:', newhdu.data.min(), 'to', newhdu.data.max())
    return newhdu

def fix_sdss_idr_file(infile, outfile):
    print('Reading', infile)
    newhdu = fix_sdss_idr(pyfits.open(infile)[0])
    print('Writing', outfile)
    pyfits_wireto(newhdu, outfile)


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print('Usage: %s <input-fits-file> <output-fits-file>' % sys.argv[0])
        sys.exit(-1)
    infile = sys.argv[1]
    outfile = sys.argv[2]

    fix_sdss_idr_file(infile, outfile)
    sys.exit(0)
